/* ANALYZER */



declare v(255) character, (left_part, right_head) (255) bit(8),
   production(255) bit(32), (on_left, on_right) (255) bit(1);
declare index(255) bit(8), (ind, ind1) (255) bit(16);
declare sort#(255) fixed, ambiguous bit(1);
declare (nsy, npr, sp, cp, nt, level, goal_symbol) fixed;
declare control(255) bit(1);
declare true literally '1', false literally '0';
declare carriage literally '1', punch literally '2', disk literally '3',
   page character initial('1'), double character initial('0'),
   eject_page literally 'output(1) = page',
   double_space literally 'output(1) = double';
declare dollar literally 'byte(''$'')', blank literally 'byte('' '')';
declare change bit(1), work("4000") bit(8);
declare head(255) fixed, tail(255) bit(8);
declare stacklimit literally '200', textlimit literally '255',
   depth literally '255'; /* must be at least 255. */
declare stack(stacklimit) bit(8), text(textlimit) bit(8),
   token_save(depth) bit(8), token fixed, mp_save(depth) bit(8),
   mp fixed, tp_save(depth) bit(8), tp fixed, p_save(depth) bit(8),
   p fixed;
declare head_table("2000") bit(8);
declare empty character initial('
                                         '); /* 80 */
   /* The image of a blank card. */
declare half_line character initial('
                               '), /* 66 */
   x12 character initial('            '); /* 12 */
declare cardimage character, outcard character, s character,
   t character;
declare (netry, first_time, last_time, this_time) fixed;
declare count(3) fixed;
declare print(3) character initial(' ','Y','N','#');
declare dots character initial(' ... ');
declare value(1) fixed initial(2, 1);
declare (i, j, k, l, m) fixed;
declare error_count fixed;
declare terminator fixed;
declare maxnf11 literally '5000', maxntrip literally '1000';
declare f11(maxnf11) fixed, nf11 fixed;
declare triple(maxntrip) fixed, tv(maxntrip) bit(2), ntrip fixed;
declare stacking bit(1) initial(true);
   /* Controls batching of grammars. */
declare maxtrouble literally '50', trouble1(maxtrouble) bit(8),
   trouble2(maxtrouble) bit(8), trouble_count fixed;
declare (basic_nsy, basic_npr) fixed;
declare iteration_count fixed;



IS_HEAD:
   procedure (i, j) bit(1);

   /* This procedure decodes the packed head table. True is v(j) is
      a head of v(i). */

   /* We must simulate a double subscript for the array
      head_table(0:255,0:255) bit(1). */

   declare (i, j) fixed;

   return 1 & shr(head_table(shl(i,5) + shr(j,3)), j & 7);

end IS_HEAD;



SET_HEAD:
   procedure (i, j);

   /* This procedure adds v(j) as a head of v(i) in head_table. */

   declare (i, j, k, l) fixed;

   change = true;
   k = shl(i, 5) + shr(j, 3);
   l = shl(1, j & 7);
   head_table(k) = head_table(k) | l;

end SET_HEAD;



CLEAR_HEADS:
   procedure;

   declare i fixed;

   do i = 0 to "2000";
      head_table(i) = 0;
   end;

end CLEAR_HEADS;



GET:
   procedure (i, j) bit(2);

   /* This procedure decodes a 2-bit entry in the work matrix. */

   /* We must simulate a double subscript for the array
      work(0:255, 0:255) bit(2). */

   declare (i, j) fixed;

   return 3 & shr(work(shl(i,6)+shr(j,2)), shl(j & 3, 1));

end GET;



SET:
   procedure (i, j, val);

   /* This procedure or's a 2-bit val into the work matrix. */

   declare (i, j, val) fixed;
   declare (k, l) fixed;

   k = shl(i, 6) + shr(j, 2);
   l = shl(val & 3, shl(j & 3, 1));
   work(k) = work(k) | l;

end SET;



CLEAR_WORK:
   procedure;

   declare i fixed;

   do i = 0 to "4000";
      work(i) = 0;
   end;

end CLEAR_WORK;



PACK:
   procedure (b1, b2, b3, b4) fixed;

   /* This procedure has the value of the 4 bytes packed into
      a 32-bit word. */

   declare (b1, b2, b3, b4) bit(8);

   return shl(b1, 24) + shl(b2, 16) + shl(b3, 8) + b4;

end PACK;



ERROR:
   procedure (message);

   declare message character;

   output = '*** Error, ' || message;
   error_count = error_count + 1;

end ERROR;



ENTER:
   procedure (env, val);

   /* This procedure records together the 2-bit val's for each unique
      env. To assist table lookup, the env's are stored in ascending
      order. They are located by a binary search. */

   declare (env, val, i, j, k) fixed;

   netry = netry + 1; /* Count entries vs. unique entries. */
   i = 0; k = ntrip + 1;
   do while i + 1 < k; /* Binary look-up. */
      j = shr(i+k, 1);
      if triple(j) > env then k = j;
      else if triple(j) < env then i = j;
      else
         do;
            tv(j) = tv(j) | val;
            return;
         end;
   end;
   if ntrip >= maxntrip then
      do;
         call ERROR('too many triples for table.');
         ntrip = 0;
      end;
   do j = 0 to ntrip - k; /* Make room in table for new entry. */
      i = ntrip - j;
      triple(i+1) = triple(i);
      tv(i+1) = tv(i);
   end;
   ntrip = ntrip + 1;
   triple(k) = env;
   tv(k) = val;

end ENTER;



ADD_TROUBLE:
   procedure (left, right);

   declare (left, right) fixed;
   declare i fixed;

   if left > basic_nsy then return;
   if left = terminator then return;
   if trouble_count = maxtrouble then return; /* Trouble enough. */
   do i = 1 to trouble_count;
      if trouble1(i) = left then if trouble2(i) = right then return;
   end;
   trouble_count = trouble_count + 1;
   trouble1(trouble_count) = left;
   trouble2(trouble_count) = right;

end ADD_TROUBLE;



LINE_OUT:
   procedure (number, line);

   /* Number a line and print it. */

   declare number fixed, line character;
   declare n character;

   n = number; number = 6 - length(n); /* 6 = margin */
   output = substr(empty, 0, number) || n || '   ' || line;

end LINE_OUT;



BUILD_CARD:
   procedure (item);

   /* Add item to outcard and punch if card boundary exceeded. */

   declare item character;

   if length(item) + length(outcard) >= 72 then
      do;
         if control(byte('P')) then output(punch) = outcard;
         if control(byte('O')) then
            output = '--- card output ---|' || outcard;
         outcard = '      ' || item;
      end;
   else outcard = outcard || ' ' || item;

end BUILD_CARD;



PUNCH_CARD:
   procedure (item);

   /* Punch outcard and item. */

   declare item character;

   call BUILD_CARD (item);
   if control(byte('P')) then output(punch) = outcard;
   if control(byte('O')) then
      output = '--- card output ---|' || outcard;
   outcard = '  ';

end PUNCH_CARD;



PRINT_MATRIX:
   procedure (title, source);

   /* Print and label the matrix specified by source (head_table or
      work). */

   declare title character, source fixed;
   declare (i, j, k, l, m, n, bot, top, margin_size, number_across,
      wide) fixed,
      (margin, line, waste, bar, pages) character,
      digit(9) character initial('0', '1', '2', '3', '4', '5', '6', '7',
         '8', '9'),
      number_high literally '48',
      gs literally '16';

   if source = 1 then wide = nt; else wide = nsy;
   margin_size = 5;
   do i = 1 to nsy;
      if length(v(i)) >= margin_size then margin_size = length(v(i)) +1;
   end;
   margin = substr(empty, 0, margin_size);
   waste = margin || '          ';
   number_across = (122 - margin_size) / (gs + 1) * gs;
   do i = 0 to 3;
      count(i) = 0;
   end;
   m = 0;
   i = (wide - 1) / number_across + 1;
   pages = ((nsy - 1) / number_high + 1) * i;
   do i = 0 to (wide - 1) / number_across;
      bot = number_across * i + 1;
      top = number_across * (i + 1);
      if top > wide then top = wide;
      bar = substr(waste, 1) || '+';
      do l = bot to top;
         bar = bar || '-';
         if l mod gs = 0 then bar = bar || '+';
      end;
      if top mod gs ^= 0 then bar = bar || '+';
      do j = 0 to (nsy - 1) / number_high;
         /* Once per page of printout. */
         eject_page;
         m = m + 1;
         output = title || ': page ' || m || ' of ' || pages;
         double_space;
         l = 100;
         do while l > 0;
            line = waste;
            do n = bot to top;
               if n < l then line = line || ' ';
               else line = line || digit(n / l mod 10);
               if n mod gs = 0 then line = line || ' ';
            end;
            output = line;
            l = l / 10;
         end;
         output = bar;
         n = number_high * (j + 1);
         if n > nsy then n = nsy;
         do k = number_high * j + 1 to n;
            l = length(v(k));
            line = v(k) || substr(margin, l) || '|';
            do l = bot to top;
               if source ^= 0 then
                  do;
                     n = GET(k, l);
                     line = line || print(n);
                     count(n) = count(n) + 1;
                  end;
               else line = line || print(IS_HEAD(k, l));
               if l mod gs = 0 then line = line || '|';
            end;
            if top mod gs ^= 0 then line = line || '|';
            call LINE_OUT(k, line);
            if k mod gs = 0 then
               output = bar;
         end;
         if k mod gs ^= 1 then output = bar;
      end;
   end;
   double_space;
   if source ^= 0 then
      do;
         output = 'Table entries summary:';
         do i = 0 to 3;
            call LINE_OUT(count(i), print(i));
         end;
      end;

end PRINT_MATRIX;



PRINT_TRIPLES:
   procedure (title);

   /* Format and print the (2,1) triples for c1. */

   declare title character, (i, j) fixed;

   if ntrip = 0 then
      do;
         double_space;
         output = 'No triples required.';
         count(1) = 0; /* so we don't punch any. */
         return;
      end;
   eject_page;
   output = title || ':';
   double_space;
   do i = 1 to 3;
      count(i) = 0;
   end;
   do i = 1 to ntrip;
      j = triple(i);
      k = tv(i);
      if k = 3 then
         do;
            call ERROR(
'stacking decision cannot be made with (2,1) context:');
            call ADD_TROUBLE(shr(j, 16), shr(j, 8) & "FF");
         end;
      call LINE_OUT(i, print(k) || ' for ' || v(shr(j, 16)) || ' ' ||
         v(shr(j, 8) & "FF") || ' ' || v(j & "FF"));
      count(k) = count(k) + 1;
   end;
   double_space;
   output = netry || ' entries for ' || ntrip || ' triples.';
   double_space;
   output = 'Table entries summary:';
   do i = 1 to 3;
      call LINE_OUT(count(i), print(i));
   end;

end PRINT_TRIPLES;



BUILD_RIGHT_PART:
   procedure (p);

   declare (p, pr) fixed;

   pr = production(p);
   t = '';
   do while pr ^= 0;
      t = ' ' || v(pr & "FF") || t;
      pr = shr(pr, 8);
   end;
   t = v(right_head(p)) || t;

end BUILD_RIGHT_PART;



OUTPUT_PRODUCTION:
   procedure (p);

   declare p fixed;

   call BUILD_RIGHT_PART (p);
   call LINE_OUT (p, v(left_part(p)) || ' ::= ' || t);

end OUTPUT_PRODUCTION;



PRINT_TIME:
   procedure;

   /* Output elapsed times. */

   declare (i, j) fixed, t character;

   double_space;
   this_time = time;
   i = this_time - last_time;
   j = i mod 100;
   i = i / 100;
   t = 'time used was ' || i || '.';
   if j < 10 then t = t || '0';
   output = t || j || ' seconds.';
   i = this_time - first_time;
   j = i mod 100;
   i = i / 100;
   t = 'total time is ' || i || '.';
   if j < 10 then t = t || '0';
   output = t || j || ' seconds.';
   last_time = this_time;

end PRINT_TIME;



LOOK_UP:
   procedure (symbol) bit(8);

   /* Get index of symbol in v. */

   declare symbol character;
   declare j fixed;

   do j = 1 to nsy;
      if v(j) = symbol then return j;
   end;
   if j = 256 then
      do;
         call ERROR('too many symbols.');
         j = 1;
      end;
   /* Add symbol to v. */
   nsy = j;
   v(j) = symbol;
   return j;

end LOOK_UP;



EXPAND:
   procedure (f11, p);

   /* Expand production p in the context of f11. */

   declare (f11, i, j, p, oldp) fixed;
      /* oldp remembers argument p from previous call to save
         repeated effort. */

   if p ^= oldp then
      do;
         oldp = p; sp = 2;
         stack(sp) = right_head(p);
         j = production(p);
         do while j ^= 0; /* Unpack production into stack. */
            i = shr(j, 24);
            if i ^= 0 then
               do;
                  sp = sp + 1;
                  stack(sp) = i;
               end;
            j = shl(j, 8);
         end;
      end;
   stack(1) = shr(f11, 8) & "FF"; /* Left context. */
   stack(sp + 1) = f11 & "FF"; /* Right context. */

end EXPAND;



READ_GRAMMAR:
   procedure;

   /* Read in and list a grammar. */

   declare (p, long) fixed;



   SCAN:
      procedure bit(8);

      /* Get a token from input card image. */

      declare lp fixed, left_bracket literally 'byte(''<'')',
         right_bracket literally 'byte(''>'')', stop fixed;

      do cp = cp to long;
         if byte(cardimage,cp) ^= blank then
            do;
               lp = cp; /* Mark left boundary of symbol. */
               if byte(cardimage,cp) = left_bracket &
                  byte(cardimage,cp+1) ^= blank then
                  stop = right_bracket;
               else
                  stop = blank;
               do cp = cp + 1 to long;
                  if byte(cardimage,cp) = stop then
                     go to delimit;
               end;
               if stop ^= blank then
                  do;
                     call ERROR('unmatched bracket: <');
                     cp = cp - 1;
                     do while byte(cardimage,cp) = blank;
                        cp = cp - 1; /* Error recovery. */
                     end;
                  end;
            delimit:
               if stop ^= blank then cp = cp + 1;
                  /* Pick up the ">". */
               t = substr(cardimage, lp, cp - lp);
                  /* Pick up the symbol. */
               return LOOK_UP(t);
            end;
      end; /* End of card. */
      t = '';
      return 0;

   end SCAN;



   GET_CARD:
      procedure bit(1);

      /* Read the next card. */

      cp = 0;
      do while true;
         cardimage = input; /* Get the card. */
         long = length(cardimage) - 1;
         if long < 0 then
            do; /* End of file detected. */
               stacking = false;
               return false;
            end;
         if byte(cardimage) = dollar then
            do; /* Control card or comment. */
               if substr(cardimage,1,3) = 'EOG' then
                  return false;
               if control(byte('L')) then
                  output = cardimage;
               control(byte(cardimage,1)) = ^control(byte(cardimage,1));
            end;
         else if substr(cardimage, 0, 72) ^= substr(empty, 0, 72) then
            do;
               cardimage = substr(cardimage, 0, 72) || '        ';
               return true;
            end;
      end;

   end GET_CARD;



   SORT_V:
      procedure;

      /* Sort the vocabulary. */

      do i = 1 to nsy;
         /* Sort on:
            1) terminal vs. non-terminal
            2) length of symbol
            3) original order of occurence */
         sort#(i) = shl(on_left(i), 16) | shl(length(v(i)),8) | i;
      end;
      /* Bubble sort. */
      k, l = nsy;
      do while k <= l;
         l = 0;
         do i = 2 to k;
            l = i - 1;
            if sort#(l) > sort#(i) then
               do;
                  j = sort#(l); sort#(l) = sort#(i); sort#(i) = j;
                  t = v(l); v(l) = v(i); v(i) = t;
                  k = l;
               end;
         end;
      end;
      do i = 1 to nsy;
         /* Build a table to locate the sorted symbols of v. */
         index(sort#(i) & "FF") = i;
      end;
      nt = nsy; /* Prepare to count non-terminal symbols. */
      do while sort#(nt) > "10000"; nt = nt - 1; end;
      /* Substitute new index numbers in productions. */
      do i = 1 to npr;
         left_part(i) = index(left_part(i));
         j = index(right_head(i));
         on_right(j) = true;
         right_head(i) = j;
         l = production(i);
         do k = 0 to 3;
            j = index(shr(l,24));
            on_right(j) = true;
            l = shl(l,8) + j;
         end;
         production(i) = l;
      end;
      terminator = index(1); /* Add "_|_" to vocabulary. */
      on_right(terminator) = true;

   end SORT_V;



   PRINT_DATE:
      procedure (message, d);

      declare message character, d fixed;
      declare month(11) character initial('January','February','March',
         'April','May','June','July','August','September','October',
         'November','December'),
         days(11) fixed initial(0, 31, 60, 91, 121, 152, 182, 213, 244,
            274, 305, 335);
      declare (year, day, m) fixed;

      year = d/1000 + 1900;
      day = d mod 1000;
      if (year & 3) ^= 0 then if day > 59 then day = day + 1;
      m = 11;
      do while day <= days(m); m = m - 1; end;
      output = message || month(m) || ' ' || day - days(m) || ', ' ||
         year || '.';

   end PRINT_DATE;



   eject_page;
   call PRINT_DATE (
'Grammar analysis - Moscow Aviation Institute - ANALYZER version of ',
      date_of_generation);
   double_space;
   call PRINT_DATE ('Today is ', date);
   double_space;
   output = '            P R O D U C T I O N S';
   double_space;
   control(byte('L')) = true;
   v(0) = '<Error: token = 0>';
   nsy = 1; v(1) = '_|_';
   npr, error_count = 0;
   do i = 0 to 255; /* Clear on_left and on_right. */
      on_left(i), on_right(i) = 0;
   end;
   do while GET_CARD; /* (Watch the side effect). */
      if npr = 255 then call ERROR ('too many productions.');
      else npr = npr + 1;
      if byte(cardimage) = blank then
         left_part(npr) = left_part(npr-1);
      else
         do;
            i = SCAN; /* Left part symbol. */
            left_part(npr) = i;
            on_left(i) = true;
         end;
      i = SCAN; /* First symbol on the right. */
      if i = 0 then call ERROR ('empty right part.');
      right_head(npr) = i;
      j, p = 0;
      do j = 1 to 4;
         i = SCAN;
         if i ^= 0 then
            p = shl(p, 8) + i; /* Pack 4 to a word. */
      end;
      if substr(cardimage, cp) ^= substr(empty, cp) then
         call ERROR (
            'too many symbols in right part. Forces to discard ' ||
            substr(cardimage, cp));
      production(npr) = p;
      if control(byte('L')) then
         do; /* Unpack table and print production. */
            if left_part(npr) = left_part(npr-1) then
               do;
                  i = length(v(left_part(npr)));
                  cardimage = substr(empty, 0, i) || '   | ';
               end;
            else
               do;
                  output = '';
                  cardimage = v(left_part(npr)) || ' ::= ';
               end;
            call BUILD_RIGHT_PART (npr);
            call LINE_OUT (npr, cardimage || t);
         end;
         if control(byte('P')) then
            do; /* Unpack table and print XPL comment. */
               p = production(npr);
               outcard = '/* ' || v(left_part(npr)) || ' ::= ' ||
                  v(right_head(npr));
               do k = 1 to 4;
                  i = shr(p, shl(4-k, 3)) & "FF";
                  if i ^= 0 then call BUILD_CARD (v(i));
               end;
               call PUNCH_CARD ('   */');
            end;
      end;
      call PRINT_TIME;
      call SORT_V;
      eject_page;
      output = '   T E R M I N A L   S Y M B O L S
             N O N T E R M I N A L S';
      double_space;
      if nsy - nt > nt then l = nsy - nt; else l = nt;
         /* l = number of lines */
      do i = 1 to l; /* Print v. */
         if i > nt then cardimage = half_line;
         else
            do; /* Terminal symbols. */
               t = i;
               j = 5 - length(t);
               cardimage = substr(substr(empty, 0, j) || t ||
                  '   ' || v(i) || half_line, 0, 66);
            end;
         k = i + nt;
         if k <= nsy then
            do; /* Non-terminal symbols. */
               t = k;
               j = 5 - length(t);
               cardimage = cardimage || substr(empty, 0, j) || t ||
                  '   ' || v(k);
            end;
         output = cardimage;
      end;
      double_space;
      goal_symbol = 0;
      do i = 1 to nsy; /* Locate the goal symbol. */
         if ^ on_right(i) then
            if goal_symbol = 0 then
               do;
                  goal_symbol = i;
                  output = v(i) || ' is the goal symbol.';
               end;
            else output = 'Another goal: ' || v(i) ||
               ' (will not be used).';
      end;
      if goal_symbol = 0 then
         do;
            goal_symbol = left_part(1);
            output = 'No goal symbol found. ' || v(goal_symbol) ||
               ' used for goal symbol.';
         end;
      basic_nsy = nsy;
      basic_npr = npr;
      trouble_count = 0;

end READ_GRAMMAR;



IMPROVE_GRAMMAR:
   procedure;

   declare (t1, t2, s1, s2) fixed, change bit(1),
      internal bit(1),
      add_on character initial ('0123456789abcdefghijkl');

   eject_page;
   output = 'Grammar modification to attempt to resolve conflicts:';
   do i = 1 to basic_nsy;
      index(i) = 0;
   end;
   do i = 1 to trouble_count; /* Step through problems. */
      double_space;
      t1 = trouble1(i); t2 = trouble2(i);
      do p = 1 to basic_npr;
         internal, change = false;
         s1 = right_head(p);
         m = production(p);
         do l = 1 to 4; /* Step through right part. */
            s2 = shr(m, 24);
            m = shl(m, 8) + s2;
            if s2 ^= 0 then
               do;
                  if s1 = t1 & IS_HEAD(s2, t2) then
                     do;
                        j, index(t1) = index(t1) + 1;
                        if nsy < 255 then nsy = nsy + 1;
                        else call ERROR ('too many symbols.');
                        s = substr(add_on, j, 1);
                        v(nsy) = '<' || v(t1) || s || '>';
                        if npr < 255 then npr = npr + 1;
                        else call ERROR ('too many productions.');
                        left_part(npr) = nsy;
                        right_head(npr) = t1;
                        production(npr) = 0;
                        change = true;
                        call OUTPUT_PRODUCTION (npr);
                        if internal then
                           m = m & "FFFF00FF" | shl(nsy, 8);
                        else
                           right_head(p) = nsy;
                     end;
                  internal = true; s1 = s2;
               end;
         end; /* of do l. */
         production(p) = m;
         if change then call OUTPUT_PRODUCTION (p);
      end; /* of do p. */
   end; /* of do i. */
   trouble_count, error_count = 0;

end IMPROVE_GRAMMAR;



COMPUTE_HEADS:
   procedure;

   /* Set up head symbol matrix. */

   call CLEAR_HEADS;
   call CLEAR_WORK;
   do i = 1 to npr; /* First get immediate heads. */
      call SET_HEAD (left_part(i), right_head(i));
   end;
   do while change; /* then compute transitive completion. */
      change = false;
      do i = nt + 1 to nsy;
         do j = nt + 1 to nsy;
            if IS_HEAD (i, j) then
               do k = 1 to nsy;
                  if IS_HEAD (j, k) then
                     if ^ IS_HEAD (i, k) then
                        call SET_HEAD (i, k);
                           /* side effect on change. */
               end;
         end;
      end;
   end;
   change = true;
   if iteration_count = 1 then
      do;
         ambiguous = false;
         do i = 1 to npr;
            j = right_head(i); k = production(i);
            do while k ^= 0;
               j = shr(k, 24); k = shl(k, 8);
            end;
            call SET(left_part(i), j, true);
         end;
         do while change;
            change = false;
            do i = nt + 1 to nsy;
               do j = nt + 1 to nsy;
                  if GET(i, j) then
                     do k = 1 to nsy;
                        if GET(j, k) then
                           if ^ GET(i, k) then
                              do;
                                 call SET(i, k, true);
                                 change = true;
                              end;
                     end;
               end;
            end;
         end;
         do i = nt + 1 to nsy;
            if IS_HEAD(i, i) then if GET(i, i) then
               do;
                  ambiguous = true;
                  call ERROR (
'grammar is ambiguous. It is left and right recursive in the symbol '
                     || v(i));
               end;
         end;
      end;
   do i = 0 to nsy; /* Finally, the reflexive transitive completion. */
      call SET_HEAD(i, i);
   end;
   call PRINT_MATRIX('Produced head symbols', 0);

end COMPUTE_HEADS;



PRODUCE:
   procedure;

   /* Run through the generation algorithm to compute f11. */

   declare maxlevel fixed, new bit(1);



   NEVER_BEEN_HERE:
      procedure;

      /* Record the f11. Return false if it is already in table. */

      declare (new_f11, i, j, k, nf11p1) fixed;

      netry = netry + 1;
      new_f11 = PACK(0, stack(sp), stack(sp-1), token);
      i = 0; k, nf11p1 = nf11 + 1;
      do while i + 1 < k; /* Binary look-up. */
         j = shr(i + k, 1);
         if f11(j) > new_f11 then k = j;
         else if f11(j) < new_f11 then i = j;
         else return false; /* Found it. */
      end;
      /* If we got here, we didn't find it. */
      if nf11 >= maxnf11 then
         do;
            call ERROR('f11 overflow.'); nf11=1;
         end;
      do j = 0 to nf11 - k; /* Make room to insert new entry. */
         f11(nf11p1-j) = f11(nf11-j);
      end;
      nf11 = nf11p1;
      f11(k) = new_f11;
      return true;

   end NEVER_BEEN_HERE;



   ADD_ITEM:
      procedure (item);

      declare item character;

      if length(cardimage) + length(item) > 130 then
         do;
            output = cardimage; cardimage = '  ';
         end;
      cardimage = cardimage || ' ' || item;

   end ADD_ITEM;



   PRINT_FORM:
      procedure;

      /* Print the current sentential form while tracing the
         generating algorithm. */

      cardimage = 'level ' || level || ':';
      do i = 1 to sp;
         call ADD_ITEM (v(stack(i)));
      end;
      call ADD_ITEM (' | ');
      do i = 0 to tp - 1;
         call ADD_ITEM (v(text(tp-i)));
      end;
      output = cardimage;

   end PRINT_FORM;



   APPLY_PRODUCTION:
      procedure;

      /* Perform one parse step (on stack and text) and recur. */

      level = level + 1; /* Simulate the effect of recursion. */
      if level > maxlevel then if level > depth then
         do;
            call ERROR('level overflow.');
            level = 1;
         end;
      else
         maxlevel = level;
      mp_save(level) = mp; /* Save pointer to left part of production,*/
      mp = sp;
      tp_save(level) = tp; /* save pointer into text, */
      p_save(level) = p; /* save number of production to be applied, */
      token_save(level) = token; /* save pointer into IS_HEAD array, */
      stack(sp) = right_head(p); /* and expand production into stack. */
      j = production(p);
      do while j ^= 0;
         k = shr(j, 24); j = shl(j, 8);
         if k ^= 0 then
            if sp = stacklimit then
               call ERROR('stack overflow.');
            else
               do;
                  sp = sp + 1; stack(sp) = k;
               end;
      end;
      if control(byte('T')) then
         call PRINT_FORM; /* Trace. */

   end APPLY_PRODUCTION;



   DIS_APPLY:
      procedure;

      /* Undo the pseudo-recursion, reversing the effect of
         APPLY_PRODUCTION. */

      token = token_save(level);
      p = p_save(level);
      tp = tp_save(level);
      sp = mp;
      mp = mp_save(level);
      stack(sp) = left_part(p);
      level = level - 1;

   end DIS_APPLY;



   do i = 0 to nsy; index(i) = 0; end;
   do i = 1 to npr;
      /* Make sure productions are properly grouped. */
      j = left_part(i);
      if j ^= left_part(i-1) then
         if index(j) = 0 then
            index(j) = i;
         else call ERROR('productions separated for ' || v(j) ||
            '. Production ' || i || ' will be ignored.');
   end;
   left_part(npr+1) = 0; /* Flag end of table. */
   /* Set initial sentential form to "_|_ <goal> _|_". */
   tp = 0;
   mp, sp = 1;
   stack(0), text(0) = terminator;
   stack(1) = goal_symbol;
   netry, nf11, level, maxlevel = 0;
   eject_page;
   output = 'Sentential form production:';
   double_space;
   if control(byte('T')) then call PRINT_FORM;
   /* Now comes the basic algorithm for generating the tables. */
production_loop:
   do while sp >= mp; /* Step through right part of production. */
      if stack(sp) > nt then /* Only non-terminals can expand. */
         do;
            new = false;
            i = text(tp);
            do token = 1 to nt; /* Step through terminal heads. */
               if IS_HEAD(i, token) then
                  if NEVER_BEEN_HERE then new = true;
            end;
            if new then
               do; /* Expand stack(sp) with all applicable rules: */
                  p = index(stack(sp));
                  do while left_part(p) = stack(sp);
                     call APPLY_PRODUCTION;
                     go to production_loop; /* Now down a level, */
            continue: /* and now back up a level. */
                     p = p + 1; /* Move on to next production. */
                  end;
               end;
         end;
      if tp = textlimit then call ERROR('text overflow.');
      else tp = tp + 1;
      text(tp) = stack(sp); /* Run the compiler backwards, */
      sp = sp - 1; /* unstacking as we go. */
   end;
   /* Fully expanded at this level. */
   call DIS_APPLY; /* so come up a level. */
   if level >= 0 then go to continue;
   if control(byte('T')) then double_space;
   output = 'F11 has ' || nf11 || ' elements.';
   output = 'The maximum depth of recursion was ' ||
      maxlevel || ' levels.';
   output = netry || ' sentential forms were examined.';

end PRODUCE;



INDEX_F11:
   procedure;

   /* Build an index into f11 for each production. */

   declare (y, yp, p) fixed;

   do i = 1 to npr;
      ind(i) = 0;
      ind1(i) = -1;
   end;
   y, yp = 0;
   f11(nf11 + 1) = shl(nsy+1, 16); /* Boundary condition. */
   do i = 1 to nf11+1; /* Check each f11. */
      if f11(i) >= yp then
         do; /* f11 for a new left part. */
            p = index(y);
            do while left_part(p) = y;
               /* Record end for old left part. */
               ind1(p) = i - 1;
               p = p + 1;
            end;
            y = shr(f11(i), 16); /* New left part. */
            p = index(y);
            do while left_part(p) = y;
               /* Record start for new left part. */
               ind(p) = i;
               p = p + 1;
            end;
            yp = shl(y + 1, 16); /* To compare with triple. */
         end;
   end;

end INDEX_F11;



SORT_PRODUCTIONS:
   procedure;

   /* Re-number the productions in an optimal order for C2. */

   do i = 1 to npr; p_save(i) = 0; end;
   p = 0;
   do i = 1 to npr;
      j = right_head(i); /* Convert 1 - 4 packing to 4 - 1 packing. */
      m = production(i);
      if m = 0 then
         do;
            m = j;
            j = 0;
            l = 7;
         end;
      else l = 6;
      tail(i) = m & "FF";
      m = m & "FFFFFF00";
      do while m ^= 0;
         k = shr(m, 24); m = shl(m, 8);
         if k ^= 0 then
            do;
               j = shl(j, 8) + k;
               l = l - 1;
            end;
      end;
      head(i) = j;
      /* Sort on:
         1) tail symbol of right part,
         2) length of right part,
         3) number of f11's. */
      sort#(i) = shl(tail(i), 23) + shl(l, 20) + ind1(i) - ind(i);
      index(i) = i;
   end; /* of do i. */
   /* Bubble sort productions. */
   k, l = npr;
   do while k <= l;
      l = -1;
      do i = 2 to k;
         l = i - 1;
         if sort#(l) > sort#(i) then
            do;
               j = sort#(l); sort#(l) = sort#(i); sort#(i) = j;
               j = index(l); index(l) = index(i); index(i) = j;
               k = l;
            end;
      end;
   end;
   index(npr+1) = 0;

end SORT_PRODUCTIONS;



COMPUTE_C1:
   procedure;

   declare (cx, ctrip, s1, s2, s3, tr, pr) fixed;

   call CLEAR_WORK;
   netry, ntrip, ctrip = 0;
   do cx = 0 to 2;
      /* Repeat basic loop 3 times:
         1) compute pairs,
         2) compute triples for pair conflicts,
         3) emit diagnostic for triple conflicts. */
      do p = 1 to npr; /* Step through the productions. */
         do i = ind(p) to ind1(p); /* Step through the expansion
                                      triples. */
            call EXPAND (f11(i), p);
            do j = 2 to sp; /* Step through right part of production. */
               k = value(j ^= sp);
               s1 = stack(j-1);
               s2 = stack(j);
               l = stack(j+1);
               do s3 = 1 to nt; /* Step through the heads of
                                   stack(j+1). */
                  if IS_HEAD(l, s3) then
                     do case cx;
                        /* case 0 - enter pair. */
                        call SET(s2, s3, k);
                        /* case 1 - if pair conflict, then enter
                           triple. */
                        if GET(s2, s3) = 3 then
                           call ENTER(PACK(0, s1, s2, s3), k);
                        /* case 2 - if triple conflict, emit
                           diagnostic. */
                        do;
                           tr = PACK(0, s1, s2, s3);
                           do m = 0 to ctrip;
                              if sort#(m) = tr then
                                 call ENTER(PACK(m, p, stack(1),
                                    stack(sp+1)), k);
                           end;
                        end; /* of case 2. */
                     end; /* of do case. */
               end; /* of do s3. */
            end; /* of do j. */
         end; /* of do i. */
      end; /* of do p. */
      do case cx; /* Clean up. */
         /* case 0. */
         do;
            do i = 1 to nt; /* Special relations for terminator. */
               if IS_HEAD(goal_symbol, i) then
                  call SET(terminator, i, value(true));
            end;
            call SET(goal_symbol, terminator, value(false));
            call PRINT_MATRIX('C1 matrix for stacking decision', 1);
            call PRINT_TIME;
         end;
         /* case 1. */
         do;
            call PRINT_TRIPLES('C1 triples for stacking decision');
            if count(3) = 0 | iteration_count > 1 then return;
            if ^ control(byte('I')) then
               if control(byte('P')) | control(byte('O')) then return;
            call PRINT_TIME;
            do i = 1 to ntrip;
               if tv(i) = 3 then
                  do;
                     sort#(ctrip) = triple(i);
                     ctrip = ctrip + 1;
                  end;
            end;
            ctrip = ctrip - 1;
            netry, ntrip = 0;
            eject_page;
            output = 'Analysis of (2,1) conflicts:';
         end;
         /* case 2. */
         do;
            j = 1;
            do m = 0 to ctrip; /* Step through conflicts. */
               do k = 0 to 1; /* Step through truth values. */
                  i = sort#(m);
                  output = '';
                  output = '   The triple ' || v(shr(i,16)) || ' ' ||
                     v(shr(i,8) & "FF") || ' ' || v(i & "FF") ||
                     ' must have the value ' || print(value(k)) ||
                     ' for';
                  output = '';
                  l = shl(m+1, 24);
                  i = j;
                  s1 = 0;
                  do while triple(i) < l & i <= ntrip;
                     if (tv(i) & value(k)) ^= 0 then
                        do;
                           tr = triple(i);
                           p = shr(tr, 16) & "FF";
                           if p ^= s1 then
                              call OUTPUT_PRODUCTION(p);
                           s1 = p;
                           output = '         in the context ' ||
                              v(shr(tr, 8) & "FF") || dots ||
                              v(tr & "FF");
                        end;
                     i = i + 1;
                  end; /* of do while. */
               end; /* of do k. */
               j = i;
               output = '';
            end; /* of do m. */
         end;
      end; /* of do case. */
   end; /* of do cx. */

end COMPUTE_C1;



COMPUTE_C2:
   procedure;

   /* Determine what (if any) context must be checked for each
      production to augment the "longest match" rule for the production
      selection function C2. */

   declare (ij, ik, tj, tk, pj, pk, jct, kct, jcl, jcr) fixed,
      proper bit(1);
   declare context_class(3) character initial ('either (0,1) or (1,0)',
      '(0,1)', '(1,0)', '(1,1)');

   eject_page;
   output = 'Context check for equal and embedded right parts:';
   do i = 1 to npr; mp_save(i), p_save(i) = 0; end;
   do j = 1 to npr - 1;
      ij = index(j);
      k = j + 1;
      ik = index(k);
      do while tail(ij) = tail(ik); /* Check all productions with
                                       the same tail symbol. */
         tj = head(ij);
         tk = head(ik);
         do while (tj & "FF") = (tk & "FF") & tj ^= 0;
            tj = shr(tj, 8); tk = shr(tk, 8);
         end;
         if tk = 0 then
            do; /* Production ik is included in ij. */
               output = '';
               output = '   There are ' || ind1(ij) - ind(ij) + 1 ||
                  ' and ' || ind1(ik) - ind(ik) + 1 ||
                  ' valid contexts, respectively, for';
               call OUTPUT_PRODUCTION (ij);
               call OUTPUT_PRODUCTION (ik);
               proper = tj ^= 0; /* ik is a proper substring. */
               if proper then
                  do;
                     jcl = shl(tj & "FF", 8);
                     do i = 1 to nsy;
                        on_right(i) = false;
                     end;
                  end;
               else p_save(ij) = 1; /* Remember that equal right parts
                                       must be distinguished somehow. */
               mp = 0;
               do pj = ind(ij) to ind1(ij);
                  jct = f11(pj) & "FFFF";
                  jcr = jct & "00FF";
                  if proper then jct = jcl | jcr;
                  else jcl = jct & "FF00";
                  do pk = ind(ik) to ind1(ik);
                     kct = f11(pk) & "FFFF";
                     if kct = jct then
                        do;
                           if mp < 4 then call ERROR (
'these productions cannot be distinguished with (1,1) context.');
                           mp = mp | 4;
                           if proper then
                              do;
                                 if ^ on_right(jcr) then
                                    output = '   ' || v(left_part(ik))
                                       || ' has ' || v(shr(jcl, 8)) ||
                                       dots || v(jcr) ||
' as context and ' || v(jcr) || ' is valid right context for ' ||
                                       v(left_part(ij));
                                 on_right(jcr) = true;
                              end;
                           else output =
'   They have equal right part and the common context '
                              || v(shr(jcl, 8)) || dots || v(jcr);
                           call ADD_TROUBLE (shr(kct,8), left_part(ik));
                        end;
                     else if (kct & "FF00") = jct then
                        mp = mp | 1;/* We can't tell by left context.*/
                     else if (kct & "00FF") = jcr then
                        mp = mp | 2;/* We can't tell by right context.*/
                  end; /* of do pk. */
               end; /* of do pj. */
               if mp < 4 then
                  do; /* Resolvable by context. */
                     if proper & (^mp) then /* Context implicit in
                                               length. */
                        output = '   They can be resolved by length.';
                     else
                        do;
                           mp_save(ij) = mp_save(ij) | mp;
                           output = '   They can be resolved by ' ||
                              context_class(mp) || ' context.';
                        end;
                  end;
            end; /* of tk = 0. */
         k = k + 1;
         ik = index(k);
      end; /* of do while. */
   end; /* of do j. */
   eject_page;
   output = 'C2 production choice function:';
   tk = 0;
   do j = 1 to npr;
      ij = index(j);
      tj = tail(ij);
      if tj ^= tk then
         do;
            tk = tj;
            double_space;
            output = '   ' || v(tj) ||
' as stack top will cause productions to be checked in this order:';
         end;
      output = '';
      call OUTPUT_PRODUCTION (ij);
      do case mp_save(ij) & 3;
         /* case 0. */
         if p_save(ij) then go to case_1;
            /* Equal right part must check. */
         else output = '         There will be no context check.';
         /* case 1. */
      case_1:
         do;
            output =
'         (0,1) context will be checked. Legal right context:';
            do i = 1 to nsy; on_right(i) = false; end;
            do pj = ind(ij) to ind1(ij);
               jcr = f11(pj) & "FF";
               if ^ on_right(jcr) then
                  do;
                     on_right(jcr) = true;
                     output = x12 || dots || v(jcr);
                  end;
            end;
         end;
         /* case 2. */
         do;
            output =
'         (1,0) context will be checked. Legal left context:';
            do i = 1 to nsy; on_left(i) = false; end;
            do pj = ind(ij) to ind1(ij);
               jcl = shr(f11(pj) & "FF00", 8);
               if ^ on_left(jcl) then
                  do;
                     on_left(jcl) = true;
                     output = x12 || v(jcl) || dots;
                  end;
            end;
         end;
         /* case 3. */
         do;
            output =
'         (1,1) context will be checked. Legal context:';
            do pj = ind(ij) to ind1(ij);
               output = x12 || v(shr(f11(pj) & "FF00", 8)) ||
                  dots || v(f11(pj) & "FF");
            end;
         end;
      end; /* of do case. */
   end; /* of do j. */

end COMPUTE_C2;



PUNCH_PRODUCTIONS:
   procedure;

   declare wide fixed;

   if ^ (control(byte('P')) | control(byte('O'))) then return;
   if control(byte('O')) then eject_page;
   outcard = '  ';
   call PUNCH_CARD('declare nsy literally ''' || nsy ||
      ''', nt literally ''' || nt || ''';');
   call BUILD_CARD('declare v(nsy) character initial(');
   do i = 0 to nsy;
      s = v(i); t = '';
      l = length(s) - 1;
      do while byte(s, l) = blank; l = l - 1; end;
      if i > basic_nsy then
         do; /* Created symbol: adjust print name. */
            l = l - 3;
            s = substr(s, 1);
         end;
      do j = 0 to l;
         if substr(s, j, 1) = '''' then t = t || '''''';
         else t = t || substr(s, j, 1);
      end;
      if i < nsy then
         call BUILD_CARD('''' || t || ''',');
      else
         call PUNCH_CARD('''' || t || ''');');
   end;
   l = length(v(nt));
   call BUILD_CARD('declare v_index(' || l || ') bit(8) initial(');
   j = 1;
   do i = 1 to l;
      call BUILD_CARD(j || ',');
      do while length(v(j)) = i; j = j + 1; end;
   end;
   call PUNCH_CARD(nt + 1 || ');');
   if nt <= 15 then wide = 16; /* Force long bit strings. */
   else wide = nt;
   i = 2 * wide + 2;
   call PUNCH_CARD('declare c1(nsy) bit(' || i || ') initial(');
   do i = 0 to nsy;
      t = '   "(2)';
      do j = 0 to wide;
         if j mod 5 = 0 then
            do;
               call BUILD_CARD(t);
               t = '';
            end;
         t = t || GET(i, j);
      end;
      if i < nsy then call PUNCH_CARD(t || '",');
      else call PUNCH_CARD(t || '");');
   end;
   k = count(1) - 1;
   if k < 0 then
      do;
         call PUNCH_CARD('declare nc1triples literally ''0'';');
         call PUNCH_CARD('declare c1triples(0) fixed;');
      end;
   else
      do;
         call PUNCH_CARD(
            'declare nc1triples literally ''' || k || ''';');
         call BUILD_CARD(
            'declare c1triples(nc1triples) fixed initial(');
         j = 0;
         do i = 1 to ntrip;
            if tv(i) = 1 then
               do;
                  if j = k then
                     call PUNCH_CARD(triple(i) || ');');
                  else
                     call BUILD_CARD(triple(i) || ',');
                  j = j + 1;
               end;
         end;
      end;
   call BUILD_CARD('declare prtb(' || npr || ') fixed initial( 0,');
   do i = 1 to npr - 1;
      call BUILD_CARD(head(index(i)) || ',');
   end;
   call PUNCH_CARD(head(index(npr)) || ');');
   call BUILD_CARD('declare prdtb(' || npr || ') bit(8) initial( 0,');
   do i = 1 to npr;
      l = index(i);
      if l > basic_npr then l = 0;
      if i < npr then call BUILD_CARD(l || ',');
      else call PUNCH_CARD(l || ');');
   end;
   call BUILD_CARD('declare hdtb(' || npr || ') bit(8) initial( 0,');
   do i = 1 to npr - 1;
      call BUILD_CARD(left_part(index(i)) || ',');
   end;
   call PUNCH_CARD(left_part(index(npr)) || ');');
   call BUILD_CARD('declare prlength(' || npr|| ') bit(8) initial( 0,');
   do i = 1 to npr;
      j = 1; k = head(index(i));
      do while k ^= 0;
         j = j + 1; k = shr(k, 8);
      end;
      if i = npr then call PUNCH_CARD(j || ');');
      else call BUILD_CARD(j || ',');
   end;
   call BUILD_CARD(
      'declare context_case(' || npr || ') bit(8) initial( 0,');
   do i = 1 to nsy; tp_save(i) = 0; end;
   do i = 1 to npr; /* Compute context case. */
      j = mp_save(index(i)); /* Set up in COMPUTE_C2, used here. */
      k = left_part(index(i));
      do case j;
         /* case 0: can tell by either left or right context,
                    use length to decide unless equal. */
         j = p_save(index(i)); /* Use the chop test, if required. */
         /* case 1: use C1 matrix for this one. */
         ;
         /* case 2: need left context table. */
         tp_save(k) = tp_save(k) | 1;
         /* case 3: need both left and right context. */
         tp_save(k) = tp_save(k) | 2;
      end;
      token_save(k) = index(i);
      if i = npr then call PUNCH_CARD(j || ');');
      else call BUILD_CARD(j || ',');
   end;
   j = 0; /* Construct context tables for C2. */
   do i = nt + 1 to nsy; /* Step through non-terminals. */
      sort#(i) = j;
      if tp_save(i) then
         do;
            k = token_save(i); m = 0;
            do l = ind(k) to ind1(k);
               p = shr(f11(l), 8) & "FF";
               if p ^= m then
                  do;
                     work(j), m = p;
                     j = j + 1; /* Count the number of entries. */
                  end;
            end;
         end;
   end;
   if j = 0 then j = 1; /* Assure non-negative upper bound for array. */
   call BUILD_CARD(
      'declare left_context(' || j - 1 || ') bit(8) initial(');
   do i = 0 to j - 2;
      call BUILD_CARD(work(i) || ',');
   end;
   call PUNCH_CARD(work(j-1) || ');');
   if j > 255 then k = 16; else k =8; /* j < 256 allows 8 bit packing.*/
   call BUILD_CARD('declare left_index(' || nsy - nt || ') bit(' || k ||
      ') initial(');
   do i = nt + 1 to nsy;
      call BUILD_CARD(sort#(i) || ',');
   end;
   call PUNCH_CARD(j || ');');
   j = 0;
   do i = nt + 1 to nsy;
      sort#(i) = j; /* Record where each non-terminal starts. */
      if shr(tp_save(i), 1) then /* We need both contexts. */
         do;
            k = token_save(i);
            do l = ind(k) to ind1(k);
               triple(j) = f11(l) & "FFFF"; j = l + 1;
            end;
         end;
   end;
   if j = 0 then j = 1; /* Assure non-negative upper bound for array. */
   call BUILD_CARD(
      'declare context_triple(' || j - 1 || ') fixed initial(');
   do i = 0 to j - 2;
      call BUILD_CARD(triple(i) || ',');
   end;
   call PUNCH_CARD(triple(j-1) || ');');
   if j > 255 then k = 16; else k =8; /* j < 256 allows 8 bit packing.*/
   call BUILD_CARD('declare triple_index(' || nsy - nt || ') bit(' ||
      k || ') initial(');
   do i = nt + 1 to nsy; /* Punch marginal index table. */
      call BUILD_CARD(sort#(i) || ',');
   end;
   call PUNCH_CARD(j || ');');
   do i = 0 to nsy; p_save(i) = 0; end;
   do i = 1 to npr; /* Step through the productions. */
      p = tail(index(i)); /* Marginal index into production table. */
      if p_save(p) = 0 then p_save(p) = i;
   end;
   p_save(nsy+1) = npr + 1; /* Mark the end of the production table. */
   do j = 0 to nsy - 1;
      /* Take care of symbols that never end a production. */
      i = nsy - j;
      if p_save(i) = 0 then p_save(i) = p_save(i+1);
   end;
   call BUILD_CARD('declare pr_index(' || nsy || ') bit(8) initial(');
   do i = 1 to nsy;
      call BUILD_CARD(p_save(i) || ',');
   end;
   call PUNCH_CARD(npr + 1 || ');');
   call PRINT_TIME;

end PUNCH_PRODUCTIONS;



/* The actual execution of ANALYZER is controlled from this loop. */

do while stacking;
   first_time, last_time = time;
   iteration_count, trouble_count = 1;
   do while trouble_count > 0;
      if iteration_count = 1 then call READ_GRAMMAR;
      else call IMPROVE_GRAMMAR;
      call PRINT_TIME;
      call COMPUTE_HEADS;
      call PRINT_TIME;
      if ^ ambiguous then
         do;
            call PRODUCE;
            call PRINT_TIME;
            call INDEX_F11;
            call SORT_PRODUCTIONS;
            call COMPUTE_C1;
            call PRINT_TIME;
            call COMPUTE_C2;
            call PRINT_TIME;
            double_space;
         end;
      output = 'Analysis complete for iteration ' || iteration_count;
      if error_count = 0 then
         output = 'No errors were detected.';
      else if error_count = 1 then
         output = '* One error was detected.';
      else if error_count <= 20 then
         output = substr('********************', 0, error_count) ||
            ' ' || error_count || ' errors were detected.';
      else
         output = '******************** ' || error_count ||
            ' errors were detected.';
      iteration_count = iteration_count + 1;
      if ambiguous | ^ control(byte('I')) then trouble_count = 0;
   end;
   if ^ ambiguous then call PUNCH_PRODUCTIONS;
   if control(byte('P')) then output = 'Punching complete.';
end;
return error_count;



eof
