/*
 * Emulator for the IBM S/360
 *
 * Copyright 1984, 1996, 1997, David E. Bodenstab
 * All rights reserved.
 */

#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h>
#include "s370.h"

typedef	enum{ FALSE=0, TRUE=1 } boolean;

unsigned char	*_ZERO, *_MAX;
int		_SIZE, *_REGS;
struct psw	*_PSW;
boolean		_EXECUTING = FALSE;
boolean		T = FALSE;

char		*strnhex( unsigned char*, unsigned char*, unsigned );
int		al( unsigned int*, int );
void		d( int*, int );
void		dr( int*, int* );
void		m( int*, int );
void		mr( int*, int* );
void		svc( int, struct psw*, int* );
void		slda( int*, int );
void		sldl( int*, int );
void		srda( int*, int );
void		srdl( int*, int );

#ifdef TRACE
struct	iloop
	{
	struct iloop	*next;
	unsigned char	*ia;
	}
	itrace[10] ={	{itrace+1}, {itrace+2}, {itrace+3}, {itrace+4}, {itrace+5},
			{itrace+6}, {itrace+7}, {itrace+8}, {itrace+9}, {itrace+0} }, 
	*itracep = itrace;
#endif


/*
 * Initialize the machine state
 */
 void
s370init( zero, size, psw, regs )
	unsigned char *zero;
	struct psw *psw;
	int *regs;
	{
	_MAX = (_ZERO = zero) + (_SIZE = size);
	_PSW = psw;
	_REGS = regs;
	}


/*
 * Store the word value WORD at ADDR
 */
 void
storew( addr, word )
	register struct word *addr;
	{
	register char *t = (char*)&word;
	addr = (struct word *)((unsigned long)addr & 0x00FFFFFF);
#ifdef TRACE
	if ( _EXECUTING )
		if ( (unsigned char*)addr < _ZERO || (unsigned char*)addr >= _MAX )
			abend( 0x000C5000 );
#endif
	addr->w[0] = t[3];
	addr->w[1] = t[2];
	addr->w[2] = t[1];
	addr->w[3] = t[0];
	}

/*
 * Store the halfword value WORD at ADDR
 */
 void
storeh( addr, word )
	register struct half *addr;
	{
	register char *t = (char*)&word;
	addr = (struct half *)((unsigned long)addr & 0x00FFFFFF);
#ifdef TRACE
	if ( _EXECUTING )
		if ( (unsigned char*)addr < _ZERO || (unsigned char*)addr >= _MAX )
			abend( 0x000C5000 );
#endif
	addr->h[0] = t[1];
	addr->h[1] = t[0];
	}

/*
 * Fetch the word value from ADDR
 */
 int
fetchw( addr )
	register struct word *addr;
	{
	int word;
	register char *t = (char*)&word;
	addr = (struct word *)((unsigned long)addr & 0x00FFFFFF);
#ifdef TRACE
	if ( _EXECUTING )
		if ( (unsigned char*)addr < _ZERO || (unsigned char*)addr >= _MAX )
			abend( 0x000C5000 );
#endif
	t[3] = addr->w[0];
	t[2] = addr->w[1];
	t[1] = addr->w[2];
	t[0] = addr->w[3];
	return( word );
	}

/*
 * Store the halfword value from ADDR
 */
 int
fetchh( addr )
	register struct half *addr;
	{
	short word;
	register char *t = (char*)&word;
	addr = (struct half *)((unsigned long)addr & 0x00FFFFFF);
#ifdef TRACE
	if ( _EXECUTING )
		if ( (unsigned char*)addr < _ZERO || (unsigned char*)addr >= _MAX )
			abend( 0x000C5000 );
#endif
	t[1] = addr->h[0];
	t[0] = addr->h[1];
	return( word );
	}

/*
 * Format the ASCII string P, converting non-printable characters to '.'
 */
 unsigned char *
pascii( p )
	register unsigned char *p;
	{
	static unsigned char b[17], tr[] ={ 
	"................................ ...............0123456789......"
	".ABCDEFGHIJKLMNOPQRSTUVWXYZ......abcdefghijklmnopqrstuvwxyz....."
	"................................................................"
	"................................................................"
	};

	register unsigned char *q;

	for( q=b; q<b+16; ++q )
		*q = tr[ *p++ ];

	*q = 0;

	return( b );
	}


/*
 * Format the EBCDIC string P, converting non-printable characters to '.'
 */
 unsigned char *
pebcdic( p )
	register unsigned char *p;
	{
	static unsigned char b[17], tr[] ={ 
	"................................................................"
	" ..............................................................."
	".abcdefghi.......jklmnopqr........stuvwxyz......................"
	".ABCDEFGHI.......JKLMNOPQR........STUVWXYZ......0123456789......"
	};

	register unsigned char *q;

	for( q=b; q<b+16; ++q )
		*q = tr[ *p++ ];

	*q = 0;

	return( b );
	}

/*
 * Abend routine
 *
 * Print the machine state and optionally dump memory
 */
 void
abend( code )
	{
	boolean same, first;
	unsigned char *p, b[100], x1[20], x2[20], x3[20], x4[20];
	int x, s;
	static unsigned char ho[]={ 0,10,12,14,17,19,21,23,26,28,30,32,35,37,39,41 };
	static unsigned char ao[]={ 0,46,47,48,49,50,51,52,53,54,55,56,57,58,59,60 };
	static unsigned char eo[]={ 0,63,64,65,66,67,68,69,70,71,72,73,74,75,76,77 };


	fprintf( stderr, "\nABEND S%3.3X U%4.4d\n\n", (code>>12)&0xFFF, code&0xFFF );

	if ( _PSW )
		{
		fprintf( stderr, "PSW   0001%4.4X %X0%6.6X\n\n", _PSW->ic, _PSW->ilc*2+_PSW->cc, (unsigned int) _PSW->ia );

		for( x=0; x<16; ++x )
			{
			fprintf( stderr, "R%2.2d %8.8X  ", x, _REGS[x] );
			if ( x%4 == 3 )
				putc( '\n', stderr );
			}
		putc( '\n', stderr );
		}
	else
		fprintf( stderr, "S370 uninitialized\n\n" );

#ifdef TRACE
	{
	struct iloop *ip;

	fputs( "INSTRUCTION TRACE   ", stderr );
	ip = itracep;
	first = TRUE;
	do	{
		if ( ip->ia != 0 )
			{
			if ( ! first )
				fputs( "                    ", stderr);
			first = FALSE;
			if ( (unsigned char*)ip->ia < _ZERO || (unsigned char*)ip->ia >= _MAX || (int)ip->ia & 0x01 )
				fprintf( stderr, "%6.6X  ****\n", ip->ia );
			else
				fprintf( stderr, "%6.6X  %s %s %s\n", ip->ia,
					strnhex(x1,ip->ia,2), strnhex(x2,ip->ia+2,2), strnhex(x3,ip->ia+4,2) );
			}
		}
		while( (ip=ip->next) != itracep );
	}
	putc( '\n', stderr );

	if ( T && code == 0x000C5000 )
		{
		fflush( stderr );
		return;
		}
#endif

	_EXECUTING = FALSE;

	for( x=1; x<NSIG; ++x )
		signal( x, SIG_DFL );

	p = _ZERO;
	s = _SIZE;

	fprintf( stderr, "\nMEMORY  %8.8X-%8.8X  (%dK)\n\n", (unsigned int) p, (unsigned int) p+_SIZE, _SIZE/1024 );

	fputs( " ADDR   0....... 4....... 8....... C.......  ------ASCII----- -----EBCDIC-----\n", stderr );

	if ( (x = ((int)p & 0x0000000F)) )
		{
		p = (unsigned char*)((int)p & 0xFFFFFFF0);
		sprintf( b, "%6.6X  %s %s %s %s *%s*%s*\n",
				(unsigned int) p, strnhex(x1,p,4), strnhex(x2,p+4,4), strnhex(x3,p+8,4), strnhex(x4,p+12,4),
				pascii(p), pebcdic(p) );
		memset( b+8, ' ', ho[x]-8 );
		memset( b+45, ' ', x );
		memset( b+62, ' ', x );
		fputs( b, stderr );
		p += 16;
		s -= 16-x;
		}

	for( first=TRUE,same=FALSE; s>=16; p+=16,s-=16 )
		{
		if ( ! first && 0 == memcmp(p-16,p,16) )
			{
			if ( ! same )
				{
				fprintf( stderr, "%6.6X  --- same ---\n", (unsigned int) p );
				same = TRUE;
				}
			continue;
			}
		first = same = FALSE;
		fprintf( stderr, "%6.6X  %s %s %s %s *%s*%s*\n",
				(unsigned int) p, strnhex(x1,p,4), strnhex(x2,p+4,4), strnhex(x3,p+8,4), strnhex(x4,p+12,4),
				pascii(p), pebcdic(p) );
		}

	if ( s > 0 )
		{
		sprintf( b, "%6.6X  %s %s %s %s *%s*%s*\n",
				(unsigned int) p, strnhex(x1,p,4), strnhex(x2,p+4,4), strnhex(x3,p+8,4), strnhex(x4,p+12,4),
				pascii(p), pebcdic(p) );
		memset( b+ho[s], ' ', 43-ho[s] );
		memset( b+ao[s], ' ', 61-ao[s] );
		memset( b+eo[s], ' ', 78-eo[s] );
		fputs( b, stderr );
		}

	fflush( stderr );

	abort();
	}



/*
 * S/370 op codes
 */
enum	opcodes
	{
	BALR=0x05, BCTR,BCR,
	SVC=0x0A,
	LPR=0x10,  LNR,LTR,LCR,NR,CLR,OR,XR,LR,CR,AR,SR,MR,DR,ALR,SLR,
	STH=0x40,  LA,STC,IC,EX,BAL,BCT,BC,LH,CH,AH,SH,MH,
	CVD=0x4E,  CVB,ST,
	N=0x54,    CL,O,X,L,C,A,S,M,D,AL,SL,
	BXH=0x86,  BXLE,
	SRL=0x88,  SLL,SRA,SLA,SRDL,SLDL,SRDA,SLDA,STM,TM,MVI,
	NI=0x94,   CLI,OI,XI,LM,
	MVN=0xD1,  MVC,MVZ,NC,CLC,OC,XC,
	TR=0xDC,   TRT,
	MVO=0xF1,  PACK,UNPK
	};

#ifdef TRACE
char *mnemonic[256] ={
/*00*/	""    ,""    ,""    ,""    ,""    ,"BALR","BCTR","BCR" ,""    ,""    ,"SVC" ,""    ,""    ,""    ,""    ,""    ,
/*10*/	"LPR" ,"LNR" ,"LTR" ,"LCR" ,"NR"  ,"CLR" ,"OR"  ,"XR"  ,"LR"  ,"CR"  ,"AR"  ,"SR"  ,"MR"  ,"DR"  ,"ALR" ,"SLR" ,
/*20*/	""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,
/*30*/	""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,
/*40*/	"STH" ,"LA"  ,"STC" ,"IC"  ,"EX"  ,"BAL" ,"BCT" ,"BC"  ,"LH"  ,"CH"  ,"AH"  ,"SH"  ,"MH"  ,""    ,"CVD" ,"CVB" ,
/*50*/	"ST"  ,""    ,""    ,""    ,"N"   ,"CL"  ,"O"   ,"X"   ,"L"   ,"C"   ,"A"   ,"S"   ,"M"   ,"D"   ,"AL"  ,"SL"  ,
/*60*/	""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,
/*70*/	""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,
/*80*/	""    ,""    ,""    ,""    ,""    ,""    ,"BXH" ,"BXLE","SRL" ,"SLL" ,"SRA" ,"SLA" ,"SRDL","SLDL","SRDA","SLDA",
/*90*/	"STM" ,"TM"  ,"MVI" ,""    ,"NI"  ,"CLI" ,"OI"  ,"XI"  ,"LM"  ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,
/*A0*/	""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,
/*B0*/	""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,
/*C0*/	""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,
/*D0*/	""    ,"MVN" ,"MVC" ,"MVZ" ,"NC"  ,"CLC" ,"OC"  ,"XC"  ,""    ,""    ,""    ,""    ,"TR"  ,"TRT" ,""    ,""    ,
/*E0*/	""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,
/*F0*/	""    ,"MVO" ,"PACK","UNPK",""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,""    ,
		};
#endif

/* Only valid without using fetchh() */
struct	ILC	{ uint ILC:2, :6, :8; };
struct	RR	{ uint :8, R2:4, R1:4; };
struct	RX	{ uint :8, X2:4, R1:4; };
struct	RS	{ uint :8, R3:4, R1:4; };
struct	SI	{ uint :8, I2:8; };
struct	SS1	{ uint :8, L1:8; };
struct	SS2	{ uint :8, L2:4, L1:4; };

char _V_, _C_;	/* oVerflow and Carry flags */


/*
 * Emulate a single S/360 instruction
 */
 boolean
s370( pswp, regp )
	register struct psw *pswp;
	register int *regp;
	{
	register unsigned char *pc, *ea1, *ea2, *ea3;
	int b, x, l, l1, l2;
	enum opcodes op;
	struct BDDD bddd;

#ifdef TRACE
	unsigned char *oea1, *oea2, *oea3, *opc, x1[20], x2[20], x3[20], x4[20], h1[520], h2[520];
	int ol, ol1, ol2;
#ifdef DEBUG
	if ( T )
		{
		fprintf( stderr, "PSW   0001%4.4X %X0%6.6X\n", pswp->ic, pswp->ilc*4+pswp->cc, pswp->ia );
		for( x=0; x<16; ++x )
			{
			fprintf( stderr, "R%2.2d %8.8X  ", x, regp[x] );
			if ( x%4 == 3 )
				putc( '\n', stderr );
			}
		}
#endif

	itracep->ia = pswp->ia;
	itracep = itracep->next;

	if ( (int)pswp->ia & 0x01 )
		abend( 0x000C6000 );

	if ( _EXECUTING )
		/* turn off to allow instruction fetch for execute target */
		_EXECUTING = FALSE;
	else
		{
		if ( pswp->ia < _ZERO || pswp->ia >= _MAX )
			abend( 0x000C5000 );
		}
#endif

	/*
	 * get effective addresses
	 */
	pc = pswp->ia;
#ifdef TRACE
	opc = pc;
#endif

	l1 = l2 = 0;

	switch( op = (enum opcodes)( *pc ) )
		{
	/* RR */
	case ALR:
	case AR:
	case BALR:
	case BCTR:
	case CLR:
	case CR:
	case DR:
	case LCR:
	case LNR:
	case LPR:
	case LR:
	case LTR:
	case MR:
	case NR:
	case OR:
	case SLR:
	case SR:
	case XR:
		pswp->ilc = ((struct ILC *)pc)->ILC;
		ea1 = (unsigned char *)(regp + ((struct RR*)pc)->R1);
		ea2 = (unsigned char *)(regp + ((struct RR*)pc)->R2);
#ifdef TRACE
		if ( T )
			if ( op==MR || op==DR )
				fprintf( stderr, "%6.6X  RR  %-4s  %s .... ....  R%X=%8.8X %8.8X  R%X=%8.8X\n",
					pc, mnemonic[*pc], strnhex(x1,pc,2),
					((struct RR*)pc)->R1, *((int*)ea1), *((int*)ea1+1),
					((struct RR*)pc)->R2, *((int*)ea2) );
			else
				fprintf( stderr, "%6.6X  RR  %-4s  %s .... ....  R%X=%8.8X  R%X=%8.8X\n",
					pc, mnemonic[*pc], strnhex(x1,pc,2),
					((struct RR*)pc)->R1, *((int*)ea1),
					((struct RR*)pc)->R2, *((int*)ea2) );
#endif
		pc += 2;
		break;
	case BCR:
		pswp->ilc = ((struct ILC *)pc)->ILC;
		ea1 = (unsigned char *)((int)((struct RR*)pc)->R1);
		ea2 = (unsigned char *)(regp + ((struct RR*)pc)->R2);
#ifdef TRACE
		if ( T )
			fprintf( stderr, "%6.6X  RR  %-4s  %s .... ....  MASK=%X       R%X=%8.8X\n",
				pc, mnemonic[*pc], strnhex(x1,pc,2),
				ea1,
				((struct RR*)pc)->R2, *((int*)ea2) );
#endif
		pc += 2;
		break;
	case SVC:
		pswp->ilc = ((struct ILC *)pc)->ILC;
		ea2 = (unsigned char *)((uint)((struct SI*)pc)->I2);
#ifdef TRACE
		if ( T )
			fprintf( stderr, "%6.6X  RR  %-4s  %s .... ....  CODE=%d\n",
				pc, mnemonic[*pc], strnhex(x1,pc,2),
				ea2 );
#endif
		pc += 2;
		break;
	/* RX */
	case A:
	case AH:
	case AL:
	case BAL:
	case BCT:
	case C:
	case CH:
	case CL:
	case CVB:
	case CVD:
	case D:
	case EX:
	case IC:
	case L:
	case LA:
	case LH:
	case M:
	case MH:
	case N:
	case O:
	case S:
	case SH:
	case SL:
	case ST:
	case STC:
	case STH:
	case X:
		pswp->ilc = ((struct ILC *)pc)->ILC;
		ea1 = (unsigned char *)(regp + ((struct RX*)pc)->R1);
		*((short*)&bddd) = fetchh( pc+2 );
		if ( (b=bddd.B) )
			ea2 = (unsigned char *)(regp[b] + bddd.DDD);
		else
			ea2 = (unsigned char *)((int)bddd.DDD);
		if ( (x=((struct RX*)pc)->X2) )
			ea2 += regp[x];
#ifdef TRACE
		if ( T )
			if ( (ea3=(unsigned char*)((int)ea2 & 0x00FFFFFF)) < _ZERO || ea3 >= _MAX )
				{
				if ( op==M || op==D )
					fprintf( stderr, "%6.6X  RX  %-4s  %s %s ....  R%X=%8.8X %8.8X  OP2=%6.6X -> ********\n",
						pc, mnemonic[*pc], strnhex(x1,pc,2), strnhex(x2,pc+2,2),
						((struct RX*)pc)->R1, *((int*)ea1), *((int*)ea1+1),
						(int)ea2&0xFFFFFF );
				else
					fprintf( stderr, "%6.6X  RX  %-4s  %s %s ....  R%X=%8.8X  OP2=%6.6X -> %s\n",
						pc, mnemonic[*pc], strnhex(x1,pc,2), strnhex(x2,pc+2,2),
						((struct RX*)pc)->R1, *((int*)ea1),
						(int)ea2&0xFFFFFF, (op==LA)? "" : "********" );
				}
			else
				{
				if ( op==M || op==D )
					fprintf( stderr, "%6.6X  RX  %-4s  %s %s ....  R%X=%8.8X %8.8X  OP2=%6.6X -> %s\n",
						pc, mnemonic[*pc], strnhex(x1,pc,2), strnhex(x2,pc+2,2),
						((struct RX*)pc)->R1, *((int*)ea1), *((int*)ea1+1),
						ea3, strnhex(x3,ea3,4) );
				else
					fprintf( stderr, "%6.6X  RX  %-4s  %s %s ....  R%X=%8.8X  OP2=%6.6X -> %s\n",
						pc, mnemonic[*pc], strnhex(x1,pc,2), strnhex(x2,pc+2,2),
						((struct RX*)pc)->R1, *((int*)ea1),
						ea3, (op==LA)? (unsigned char*)"" : (unsigned char*)strnhex(x3,ea3,4) );
				}
#endif
		pc += 4;
		break;
	case BC:
		pswp->ilc = ((struct ILC *)pc)->ILC;
		ea1 = (unsigned char *)((int)((struct RX*)pc)->R1);
		*((short*)&bddd) = fetchh( pc+2 );
		if ( (b=bddd.B) )
			ea2 = (unsigned char *)(regp[b] + bddd.DDD);
		else
			ea2 = (unsigned char *)((int)bddd.DDD);
		if ( (x=((struct RX*)pc)->X2) )
			ea2 += regp[x];
#ifdef TRACE
		if ( T )
			fprintf( stderr, "%6.6X  RX  %-4s  %s %s ....  MASK=%X       OP2=%6.6X\n",
				pc, mnemonic[*pc], strnhex(x1,pc,2), strnhex(x2,pc+2,2),
				ea1,
				(int)ea2&0xFFFFFF );
#endif
		pc += 4;
		break;
	/* SS */
	case MVO:
	case PACK:
	case UNPK:
		l1 = ((struct SS2*)pc)->L1;
		l2 = ((struct SS2*)pc)->L2;
	case CLC:
	case MVC:
	case MVN:
	case MVZ:
	case NC:
	case OC:
	case TR:
	case TRT:
	case XC:
		l = ((struct SS1*)pc)->L1;
		pswp->ilc = ((struct ILC *)pc)->ILC;
		*((short*)&bddd) = fetchh( pc+2 );
		if ( (b=bddd.B) )
			ea1 = (unsigned char *)((regp[b] + bddd.DDD) & 0x00FFFFFF);
		else
			ea1 = (unsigned char *)((int)bddd.DDD);
		*((short*)&bddd) = fetchh( pc+4 );
		if ( (b=bddd.B) )
			ea2 = (unsigned char *)((regp[b] + bddd.DDD) & 0x00FFFFFF);
		else
			ea2 = (unsigned char *)((int)bddd.DDD);
#ifdef TRACE
		if ( T )
			switch( op )
				{
			case MVO:
			case PACK:
			case UNPK:
				fprintf( stderr, "%6.6X  SS  %-4s  %s %s %s  OP1=%6.6X -> %s\n                                  OP2=%6.6X -> %s\n",
					pc, mnemonic[*pc], strnhex(x1,pc,2), strnhex(x2,pc+2,2), strnhex(x3,pc+4,2),
					ea1, strnhex(h1,ea1,1+l1),
					ea2, strnhex(h2,ea2,1+l2) );
				break;
			case TR:
				fprintf( stderr, "%6.6X  SS  %-4s  %s %s %s  OP1=%6.6X -> %s\n                                  OP2=%6.6X -> %s\n",
					pc, mnemonic[*pc], strnhex(x1,pc,2), strnhex(x2,pc+2,2), strnhex(x3,pc+4,2),
					ea1, strnhex(h1,ea1,1+l),
					ea2, strnhex(h2,ea2,256) );
				break;
			case TRT:
				fprintf( stderr, "%6.6X  SS  %-4s  %s %s %s  R1=%8.8X  R2=%8.8X\n                                  OP1=%6.6X -> %s\n                                  OP2=%6.6X -> %s\n",
					pc, mnemonic[*pc], strnhex(x1,pc,2), strnhex(x2,pc+2,2), strnhex(x3,pc+4,2),
					regp[1], regp[2],
					ea1, strnhex(h1,ea1,1+l),
					ea2, strnhex(h2,ea2,256) );
				break;
			default:
				fprintf( stderr, "%6.6X  SS  %-4s  %s %s %s  OP1=%6.6X -> %s\n                                  OP2=%6.6X -> %s\n",
					pc, mnemonic[*pc], strnhex(x1,pc,2), strnhex(x2,pc+2,2), strnhex(x3,pc+4,2),
					ea1, strnhex(h1,ea1,1+l),
					ea2, strnhex(h2,ea2,1+l) );
				break;
				}
#endif
		pc += 6;
		break;
	/* SI */
	case CLI:
	case MVI:
	case NI:
	case OI:
	case TM:
	case XI:
		pswp->ilc = ((struct ILC *)pc)->ILC;
		*((short*)&bddd) = fetchh( pc+2 );
		if ( (b=bddd.B) )
			ea1 = (unsigned char *)((regp[b] + bddd.DDD) & 0x00FFFFFF);
		else
			ea1 = (unsigned char *)((int)bddd.DDD);
		ea2 = (unsigned char *)((int)((struct SI*)pc)->I2);
#ifdef TRACE
		if ( T )
			fprintf( stderr, "%6.6X  SI  %-4s  %s %s ....  OP1=%6.6X -> %s  OP2=%2.2X\n",
				pc, mnemonic[*pc], strnhex(x1,pc,2), strnhex(x2,pc+2,2),
				ea1, strnhex(x3,ea1,4),
				ea2 );
#endif
		pc += 4;
		break;
	/* RS */
	case BXH:
	case BXLE:
	case LM:
	case SLA:
	case SLDA:
	case SLDL:
	case SLL:
	case SRA:
	case SRDA:
	case SRDL:
	case SRL:
	case STM:
		pswp->ilc = ((struct ILC *)pc)->ILC;
		ea1 = (unsigned char *)((int)((struct RS*)pc)->R1);
		*((short*)&bddd) = fetchh( pc+2 );
		if ( (b=bddd.B) )
			ea2 = (unsigned char *)(regp[b] + bddd.DDD);
		else
			ea2 = (unsigned char *)((int)bddd.DDD);
		ea3 = (unsigned char *)((int)((struct RS*)pc)->R3);
#ifdef TRACE
		if ( T )
			switch( op )
				{
			case BXH:
			case BXLE:
				fprintf( stderr, "%6.6X  RS  %-4s  %s %s ....  R%X=%8.8X  R%X=%8.8X  OP2=%6.6X -> %s\n",
					pc, mnemonic[*pc], strnhex(x1,pc,2), strnhex(x2,pc+2,2),
					ea1, regp[(int)ea1],
					ea3, regp[(int)ea3],
					(int)ea2&0xFFFFFF, strnhex(x3,(char*)((int)ea2&0xFFFFFF),4) );
				break;
			case LM:
			case STM:
				fprintf( stderr, "%6.6X  RS  %-4s  %s %s ....  R%X=%8.8X  R%X=%8.8X  OP2=%6.6X -> %s\n",
					pc, mnemonic[*pc], strnhex(x1,pc,2), strnhex(x2,pc+2,2),
					ea1, regp[(int)ea1],
					ea3, regp[(int)ea3],
					(int)ea2&0xFFFFFF, strnhex(x3,(char*)((int)ea2&0xFFFFFF),4) );
				break;
			case SLA:
			case SLL:
			case SRA:
			case SRL:
				fprintf( stderr, "%6.6X  RS  %-4s  %s %s ....  R%X=%8.8X  OP2=%d\n",
					pc, mnemonic[*pc], strnhex(x1,pc,2), strnhex(x2,pc+2,2),
					ea1, regp[(int)ea1],
					(int)ea2&0x1F );
				break;
			case SRDA:
			case SRDL:
			case SLDA:
			case SLDL:
				fprintf( stderr, "%6.6X  RS  %-4s  %s %s ....  R%X=%8.8X %8.8X  OP2=%d\n",
					pc, mnemonic[*pc], strnhex(x1,pc,2), strnhex(x2,pc+2,2),
					ea1, regp[(int)ea1], regp[(int)ea1+1],
					(int)ea2&0x3F );
				break;
				}
#endif
		pc += 4;
		break;
	default:
		abend( 0x000C1000 );
		}


	/*
	 * do the operation
	 */
#ifdef TRACE
	oea1 = ea1;
	oea2 = ea2;
	oea3 = ea3;
	ol = l;
	ol1 = l1;
	ol2 = l2;
#endif

	_EXECUTING = TRUE;

	switch( op )
		{
	case ALR:
		if ( (*(int*)ea1 = al((unsigned int*)ea1,*(int*)ea2)) )
			pswp->cc = _C_? 3:1;
		else
			pswp->cc = _C_? 2:0;
		break;
	case AR:
		*(uint*)ea1 += *(uint*)ea2;
		break;
	case BALR:
		ea3 = (unsigned char*)(*((uint*)ea2));
		*(uint*)ea1 = (uint)pc;
		if ( ea2 != (unsigned char*)regp )
			pc = (unsigned char*)((uint)ea3 & 0x00FFFFFF);
		break;
	case BCTR:
		if ( (*(uint*)ea1-=1) != 0 && ea2 != (unsigned char*)regp )
			pc = (unsigned char *)((*(uint*)ea2) & 0x00FFFFFF);
		break;
	case DR:
#ifdef TRACE
		if ( *((int*)ea2) == 0 )
			abend( 0x000C9000 );
#endif
		dr( (int*) ea1, (int*) ea2 );
		break;
	case LCR:
		*(uint*)ea1 = - *(uint*)ea2;
		break;
	case LNR:
		if ( *(uint*)ea2 & 0x80000000 )
			{
			*(uint*)ea1 = *(uint*)ea2;
			pswp->cc = 1;
			}
		else
			if ( (*(uint*)ea1 = - *(uint*)ea2) )
				pswp->cc = 1;
			else
				pswp->cc = 0;
		break;
	case LPR:
		if ( *(uint*)ea2 & 0x80000000 )
			{
			*(uint*)ea1 = - *(uint*)ea2;
			pswp->cc = 2;
			}
		else
			if ( (*(uint*)ea1 = *(uint*)ea2) )
				pswp->cc = 2;
			else
				pswp->cc = 0;
		break;
	case LR:
	case LTR:
		*(uint*)ea1 = *(uint*)ea2;
		break;
	case MR:
		mr( (int*) ea1, (int*) ea2 );
		break;
	case NR:
		if ( (*(uint*)ea1 &= *(uint*)ea2) )
			pswp->cc = 1;
		else
			pswp->cc = 0;
		break;
	case OR:
		if ( (*(uint*)ea1 |= *(uint*)ea2) )
			pswp->cc = 1;
		else
			pswp->cc = 0;
		break;
	case SR:
		*(uint*)ea1 -= *(uint*)ea2;
		break;
	case SLR:
		if ( *(int*)ea2 != 0 )
			if ( (*(int*)ea1 = al((unsigned int*)ea1,-*(int*)ea2)) )
				pswp->cc = _C_? 3:1;
			else
				pswp->cc = 2;
		else
			pswp->cc = (*(int*)ea1)? 3:2;
		break;
	case XR:
		if ( (*(uint*)ea1 ^= *(uint*)ea2) )
			pswp->cc = 1;
		else
			pswp->cc = 0;
		break;
	case BCR:
		if ( (8>>pswp->cc) & (int)ea1 )
			pc = (unsigned char*)(*(unsigned long*)ea2 & 0x00FFFFFF);
		break;
	case SVC:
		svc( (int) ea2, pswp, regp );
		break;
	case A:
		*(uint*)ea1 += (uint)fetchw( ea2 );
		break;
	case AH:
		*(uint*)ea1 += (uint)fetchh( ea2 );
		break;
	case AL:
		if ( (*(int*)ea1 = al((unsigned int*)ea1,fetchw(ea2))) )
			pswp->cc = _C_? 3:1;
		else
			pswp->cc = _C_? 2:0;
		break;
	case BAL:
		*(uint*)ea1 = (uint)pc;
		pc = (unsigned char *)(((uint)ea2) & 0x00FFFFFF);
		break;
	case BCT:
		if ( (*(uint*)ea1-=1) != 0 )
			pc = (unsigned char *)(((uint)ea2) & 0x00FFFFFF);
		break;
	case D:
#ifdef TRACE
		if ( *((uint*)ea2) == 0 )
			abend( 0x000C9000 );
#endif
		d( (int*) ea1, fetchw(ea2) );
		break;
	case EX:
		{
		static unsigned char inst[6];
		struct psw tpsw;
#ifdef TRACE
		if ( (ea3=(unsigned char*)((int)ea2 & 0x00FFFFFF)) < _ZERO || ea3 >= _MAX )
			abend( 0x000C5000 );
#endif
		memcpy( inst, (char*)((int)ea2 & 0x00FFFFFF), 6 );
		if ( ea1 != (unsigned char*)regp )
			inst[1] |= *ea1;
		tpsw.cc = pswp->cc;
		tpsw.ia = inst;
		s370( &tpsw, regp );
		pswp->cc = tpsw.cc;
		}
		break;
	case IC:
#ifdef TRACE
		if ( (ea3=(unsigned char*)((int)ea2 & 0x00FFFFFF)) < _ZERO || ea3 >= _MAX )
			abend( 0x000C5000 );
#endif
		*ea1 = *(unsigned char*)((int)ea2 & 0x00FFFFFF);
		break;
	case L:
		*(uint*)ea1 = (uint)fetchw( ea2 );
		break;
	case LA:
		*(uint*)ea1 = (uint)ea2 & 0x00FFFFFF;
		break;
	case LH:
		*(uint*)ea1 = fetchh( ea2 );
		break;
	case M:
		m( (int*) ea1, fetchw(ea2) );
		break;
	case MH:
		*(int*)ea1 *= fetchh( ea2 );
		break;
	case N:
		if ( (*(uint*)ea1 &= (uint)fetchw( ea2 )) )
			pswp->cc = 1;
		else
			pswp->cc = 0;
		break;
	case O:
		if ( (*(uint*)ea1 |= (uint)fetchw( ea2 )) )
			pswp->cc = 1;
		else
			pswp->cc = 0;
		break;
	case S:
		*(uint*)ea1 -= (uint)fetchw( ea2 );
		break;
	case SH:
		*(uint*)ea1 -= fetchh( ea2 );
		break;
	case SL:
		if ( (x=fetchw(ea2)) != 0 )
			if ( (*(int*)ea1 = al((unsigned int*)ea1,-x)) )
				pswp->cc = _C_? 3:1;
			else
				pswp->cc = 2;
		else
			pswp->cc = (*(int*)ea1)? 3:2;
		break;
	case ST:
		storew( ea2, *(uint*)ea1 );
		break;
	case STC:
#ifdef TRACE
		if ( (ea3=(unsigned char*)((int)ea2 & 0x00FFFFFF)) < _ZERO || ea3 >= _MAX )
			abend( 0x000C5000 );
#endif
		*(unsigned char*)((int)ea2 & 0x00FFFFFF) = *ea1;
		break;
	case STH:
		storeh( ea2, *(uint*)ea1 );
		break;
	case X:
		if ( (*(uint*)ea1 ^= (uint)fetchw( ea2 )) )
			pswp->cc = 1;
		else
			pswp->cc = 0;
		break;
	case BC:
		if ( (8>>pswp->cc) & (int)ea1 )
			pc = (unsigned char *)(((uint)ea2) & 0x00FFFFFF);
		break;
	case MVC:
#ifdef TRACE
		if ( ea1 < _ZERO || ea1 >= _MAX )
			abend( 0x000C5000 );
		if ( ea2 < _ZERO || ea2 >= _MAX )
			abend( 0x000C5000 );
#endif
		memcpy( ea1, ea2, l+1 );
		break;
	case MVN:
#ifdef TRACE
		if ( ea1 < _ZERO || ea1 >= _MAX )
			abend( 0x000C5000 );
		if ( ea2 < _ZERO || ea2 >= _MAX )
			abend( 0x000C5000 );
#endif
		do	{
			*ea1 &= 0xF0;
			*ea1++ |= (*ea2++) & 0x0F;
			}
			while( l-- > 0 );
		break;
	case MVZ:
#ifdef TRACE
		if ( ea1 < _ZERO || ea1 >= _MAX )
			abend( 0x000C5000 );
		if ( ea2 < _ZERO || ea2 >= _MAX )
			abend( 0x000C5000 );
#endif
		do	{
			*ea1 &= 0x0F;
			*ea1++ |= (*ea2++) & 0xF0;
			}
			while( l-- > 0 );
		break;
	case NC:
#ifdef TRACE
		if ( ea1 < _ZERO || ea1 >= _MAX )
			abend( 0x000C5000 );
		if ( ea2 < _ZERO || ea2 >= _MAX )
			abend( 0x000C5000 );
#endif
		pswp->cc = 0;
		do	{
			if ( (*ea1++ &= *ea2++) )
				pswp->cc = 1;
			}
			while( l-- > 0 );
		break;
	case OC:
#ifdef TRACE
		if ( ea1 < _ZERO || ea1 >= _MAX )
			abend( 0x000C5000 );
		if ( ea2 < _ZERO || ea2 >= _MAX )
			abend( 0x000C5000 );
#endif
		pswp->cc = 0;
		do	{
			if ( (*ea1++ |= *ea2++) )
				pswp->cc = 1;
			}
			while( l-- > 0 );
		break;
	case TR:
#ifdef TRACE
		if ( ea1 < _ZERO || ea1 >= _MAX )
			abend( 0x000C5000 );
		if ( ea2 < _ZERO || ea2 >= _MAX )
			abend( 0x000C5000 );
#endif
		do	{
			*ea1 = ea2[ *ea1 ];
			++ea1;
			}
			while( l-- > 0 );
		break;
	case TRT:
#ifdef TRACE
		if ( ea1 < _ZERO || ea1 >= _MAX )
			abend( 0x000C5000 );
		if ( ea2 < _ZERO || ea2 >= _MAX )
			abend( 0x000C5000 );
#endif
		pswp->cc = 0;
		do	{
			if ( (x = ea2[ *ea1 ]) )
				{
				*((unsigned char*)&regp[2]) = x;
				regp[1] = (int)ea1;
				if ( l == 0 )
					pswp->cc = 2;
				else
					pswp->cc = 1;
				break;
				}
			++ea1;
			}
			while( l-- > 0 );
		break;
	case XC:
#ifdef TRACE
		if ( ea1 < _ZERO || ea1 >= _MAX )
			abend( 0x000C5000 );
		if ( ea2 < _ZERO || ea2 >= _MAX )
			abend( 0x000C5000 );
#endif
		pswp->cc = 0;
		do	{
			if ( (*ea1++ ^= *ea2++) )
				pswp->cc = 1;
			}
			while( l-- > 0 );
		break;
	case MVI:
#ifdef TRACE
		if ( ea1 < _ZERO || ea1 >= _MAX )
			abend( 0x000C5000 );
#endif
		*ea1 = (unsigned char)ea2;
		break;
	case NI:
#ifdef TRACE
		if ( ea1 < _ZERO || ea1 >= _MAX )
			abend( 0x000C5000 );
#endif
		if ( (*ea1 &= (unsigned char)ea2) )
			pswp->cc = 1;
		else
			pswp->cc = 0;
		break;
	case OI:
#ifdef TRACE
		if ( ea1 < _ZERO || ea1 >= _MAX )
			abend( 0x000C5000 );
#endif
		if ( (*ea1 |= (unsigned char)ea2) )
			pswp->cc = 1;
		else
			pswp->cc = 0;
		break;
	case XI:
#ifdef TRACE
		if ( ea1 < _ZERO || ea1 >= _MAX )
			abend( 0x000C5000 );
#endif
		if ( (*ea1 ^= (unsigned char)ea2) )
			pswp->cc = 1;
		else
			pswp->cc = 0;
		break;
	case SLDA:
#ifdef TRACE
		if ( (int)ea1 & 1 )
			abend( 0x000C6000 );
#endif
		slda( regp+(int)ea1, ((int)ea2 & 0x3F) );
		break;
	case SLDL:
#ifdef TRACE
		if ( (int)ea1 & 1 )
			abend( 0x000C6000 );
#endif
		sldl( regp+(int)ea1, ((int)ea2 & 0x3F) );
		break;
	case SRDA:
#ifdef TRACE
		if ( (int)ea1 & 1 )
			abend( 0x000C6000 );
#endif
		srda( regp+(int)ea1, ((int)ea2 & 0x3F) );
		break;
	case SRDL:
#ifdef TRACE
		if ( (int)ea1 & 1 )
			abend( 0x000C6000 );
#endif
		srdl( regp+(int)ea1, ((int)ea2 & 0x3F) );
		break;
	case SLA:
		regp[ (int)ea1 ] <<= ((int)ea2 & 0x1F);
		break;
	case SLL:
		((uint*)regp)[ (int)ea1 ] <<= ((int)ea2 & 0x1F);
		break;
	case SRA:
		regp[ (int)ea1 ] >>= ((int)ea2 & 0x1F);
		break;
	case SRL:
		((uint*)regp)[ (int)ea1 ] >>= ((int)ea2 & 0x1F);
		break;
	case LM:
		regp[ (int)ea1 ] = fetchw( ea2 );
		if ( ea1 != ea3 )
			do	{
				if ( (int)++ea1 > 15 )
					ea1 = (unsigned char*)0;
				regp[ (int)ea1 ] = fetchw( ea2+=4 );
				}
				while( ea1 != ea3 );
		break;
	case STM:
		storew( ea2, regp[ (int)ea1 ] );
		if ( ea1 != ea3 )
			do	{
				if ( (int)++ea1 > 15 )
					ea1 = (unsigned char*)0;
				storew( ea2+=4, regp[ (int)ea1 ] );
				}
				while( ea1 != ea3 );
		break;
	case CLR:
	case CR:
	case C:
	case CH:
	case CL:
	case CLC:
	case CLI:
	case TM:
		break;
	default:
		abend( 0x000C1000 );
		}

	/*
	 * set condition code
	 */
	switch( op )
		{
		static unsigned char zeroes[256] ={ 0 };
	case AR:
	case SR:
	case A:
	case S:
	case AH:
	case SH:
		if ( (x = *(int*)ea1) )
			if ( x < 0 )
				pswp->cc = 1;
			else
				pswp->cc = 2;
		else
			pswp->cc = 0;
		break;
	case CLR:
		if ( *(uint*)ea1 != *(uint*)ea2 )
			if ( *(uint*)ea1 < *(uint*)ea2 )
				pswp->cc = 1;
			else
				pswp->cc = 2;
		else
			pswp->cc = 0;
		break;
	case CR:
		if ( *(int*)ea1 != *(int*)ea2 )
			if ( *(int*)ea1 < *(int*)ea2 )
				pswp->cc = 1;
			else
				pswp->cc = 2;
		else
			pswp->cc = 0;
		break;
	case LTR:
		if ( (x = *(int*)ea1) )
			if ( x < 0 )
				pswp->cc = 1;
			else
				pswp->cc = 2;
		else
			pswp->cc = 0;
		break;
	case C:
		if ( *(int*)ea1 != (x=fetchw(ea2)) )
			if ( *(int*)ea1 < x )
				pswp->cc = 1;
			else
				pswp->cc = 2;
		else
			pswp->cc = 0;
		break;
	case CH:
		if ( *(int*)ea1 != (x=fetchh(ea2)) )
			if ( *(int*)ea1 < x )
				pswp->cc = 1;
			else
				pswp->cc = 2;
		else
			pswp->cc = 0;
		break;
	case CL:
		if ( *(int*)ea1 != (x=fetchw(ea2)) )
			if ( *(uint*)ea1 < (uint)x )
				pswp->cc = 1;
			else
				pswp->cc = 2;
		else
			pswp->cc = 0;
		break;
	case CLC:
#ifdef TRACE
		if ( ea1 == 0 )
			{
			if ( T )
				fprintf( stderr, "***                               OP1=0, replaced with a source of zeros\n" );
			ea1 = zeroes;
			}
		if ( ea2 == 0 )
			{
			if ( T )
				fprintf( stderr, "***                               OP2=0, replaced with a source of zeros\n" );
			ea2 = zeroes;
			}
		if ( ea1 != zeroes && (ea1 < _ZERO || ea1 >= _MAX) )
			abend( 0x000C5000 );
		if ( ea2 != zeroes && (ea2 < _ZERO || ea2 >= _MAX) )
			abend( 0x000C5000 );
#else
		if ( ea1 == 0 )
			ea1 = zeroes;
		if ( ea2 == 0 )
			ea2 = zeroes;
#endif
		pswp->cc = 0;
		do	{
			if ( *ea1 != *ea2 )
				{
				if ( *ea1 < *ea2 )
					pswp->cc = 1;
				else
					pswp->cc = 2;
				break;
				}
			++ea1;
			++ea2;
			}
			while( l-- > 0 );
		break;
	case CLI:
#ifdef TRACE
		if ( ea1 < _ZERO || ea1 >= _MAX )
			abend( 0x000C5000 );
#endif
		if ( *ea1 != (unsigned char)ea2 )
			if ( *ea1 < (unsigned char)ea2 )
				pswp->cc = 1;
			else
				pswp->cc = 2;
		else
			pswp->cc = 0;
		break;
	case TM:
#ifdef TRACE
		if ( ea1 < _ZERO || ea1 >= _MAX )
			abend( 0x000C5000 );
#endif
		if ( (x = (*ea1 & (unsigned char)ea2)) )
			if ( x == *ea1 )
				pswp->cc = 3;
			else
				pswp->cc = 1;
		else
			pswp->cc = 0;
		break;
	case SLA:
	case SRA:
		if ( (x = regp[ (int)ea1 ]) )
			if ( x < 0 )
				pswp->cc = 1;
			else
				pswp->cc = 2;
		else
			pswp->cc = 0;
		break;
	case SLDA:
	case SRDA:
		pswp->cc = _C_;
		break;
		}

	_EXECUTING = FALSE;

#ifdef TRACE
	/*
	 * show results
	 */
	switch( op )
		{
	case ALR:
	case AR:
	case LCR:
	case LNR:
	case LPR:
	case LR:
	case NR:
	case OR:
	case SLR:
	case SR:
	case XR:
		if ( T )
			fprintf( stderr, "                            CC=%d  R%X=%8.8X\n",
				pswp->cc,
				((struct RR*)opc)->R1, *((int*)oea1) );
		break;
	case BALR:
	case BCTR:
		if ( T )
			fprintf( stderr, "                                  R%X=%8.8X  PC->%6.6X\n",
				((struct RR*)opc)->R1, *((int*)oea1),
				pc );
		break;
	case DR:
	case MR:
		if ( T )
			fprintf( stderr, "                                  R%X=%8.8X %8.8X\n",
				((struct RR*)opc)->R1, *((int*)oea1), *((int*)oea1+1) );
		break;
	case BCR:
		if ( T )
			fprintf( stderr, "                                  PC->%6.6X\n",
				pc );
		break;
	case LTR:
	case CLR:
	case CR:
	case C:
	case CH:
	case CL:
	case CLC:
	case CLI:
	case TM:
		if ( T )
			fprintf( stderr, "                            CC=%d\n",
				pswp->cc );
		break;
	case D:
	case M:
		if ( T )
			fprintf( stderr, "                                  R%X=%8.8X %8.8X\n",
				((struct RX*)opc)->R1, *((int*)oea1), *((int*)oea1+1) );
		break;
	case A:
	case AH:
	case AL:
	case CVB:
	case CVD:
	case IC:
	case L:
	case LA:
	case LH:
	case MH:
	case N:
	case O:
	case S:
	case SH:
	case SL:
	case X:
		if ( T )
			fprintf( stderr, "                            CC=%d  R%X=%8.8X\n",
				pswp->cc,
				((struct RX*)opc)->R1, *((int*)oea1) );
		break;
	case ST:
	case STC:
	case STH:
		if ( T )
			fprintf( stderr, "                                               OP2=%6.6X -> %s\n",
				(int)oea2&0xFFFFFF, strnhex(x1,(char*)((int)oea2&0xFFFFFF),4) );
		break;
	case BC:
		if ( T )
			fprintf( stderr, "                                  PC->%6.6X\n",
				pc );
		break;
	case BAL:
	case BCT:
		if ( T )
			fprintf( stderr, "                                  R%X=%8.8X  PC->%6.6X\n",
				((struct RX*)opc)->R1, *((int*)oea1),
				pc );
		break;
	case MVO:
	case PACK:
	case UNPK:
		if ( T )
			fprintf( stderr, "                                  OP1=%6.6X -> %s\n",
				oea1, strnhex(h1,oea1,1+ol1) );
		break;
	case MVC:
	case MVN:
	case MVZ:
	case NC:
	case OC:
	case XC:
		if ( T )
			fprintf( stderr, "                            CC=%d  OP1=%6.6X -> %s\n",
				pswp->cc,
				oea1, strnhex(h1,oea1,1+ol) );
		break;
	case TR:
		if ( T )
			fprintf( stderr, "                                  OP1=%6.6X -> %s\n",
				oea1, strnhex(h1,oea1,1+ol) );
		break;
	case TRT:
		if ( T )
			fprintf( stderr, "                            CC=%d  R1=%8.8X  R2=%8.8X\n",
				pswp->cc,
				regp[1], regp[2] );
		break;
	case MVI:
	case NI:
	case OI:
	case XI:
		if ( T )
			fprintf( stderr, "                            CC=%d  OP1=%6.6X -> %s\n",
				pswp->cc,
				oea1, strnhex(x1,oea1,4) );
		break;
	case BXH:
	case BXLE:
		if ( T )
			fprintf( stderr, "                                  R%X=%8.8X  R%X=%8.8X  PC->%6.6X\n",
				oea1, regp[(int)oea1],
				ea3, regp[(int)ea3],
				pc );
		break;
	case SLA:
	case SLL:
	case SRA:
	case SRL:
		if ( T )
			fprintf( stderr, "                            CC=%d  R%X=%8.8X\n",
				pswp->cc,
				oea1, regp[(int)oea1] );
		break;
	case SLDA:
	case SLDL:
	case SRDA:
	case SRDL:
		if ( T )
			fprintf( stderr, "                            CC=%d  R%X=%8.8X %8.8X\n",
				pswp->cc,
				oea1, regp[(int)oea1], regp[(int)oea1+1] );
		break;
	case LM:
	case STM:
		if ( T )
			fprintf( stderr, "                                  R%X=%8.8X  R%X=%8.8X  OP2=%6.6X -> %s\n",
				oea1, regp[(int)oea1],
				ea3, regp[(int)ea3],
				(int)oea2&0xFFFFFF, strnhex(x1,(char*)((int)oea2&0xFFFFFF),4) );
		break;
		}
#endif

	pswp->ia = pc;
	return( TRUE );
	}
