.TH XPLSM 1 \" -*- nroff -*-
.SH NAME
xplsm \- XPL submonitor
.SH SYNOPSIS
.B xplsm
{\-p | \-\-program} load\-file
[{\-f | \-\-file} blocksize]
[{\-l | \-\-lines\-per\-page} linesperpage]
[{\-m | \-\-maxcore} memorysize]
[\-\-trace]
[\-\-dump]
[\-\-mmap]
[\-\-input filename]
[\-\-input2 filename]
[\-\-input3 filename]
[\-\-output filename]
[\-\-output2 filename]
[\-\-output3 filename]
[\-\-outputr4 filename]
[\-\-file1 filename]
[\-\-file2 filename]
[\-\-file3 filename]
[\-\-file4 filename]
[parms...]
.P
.B xplsm
{\-\-help, \-\-version}
.SH DESCRIPTION
.B xplsm
loads and runs an XPL load file produced by
.BR xcom ,
the XPL compiler.
The load file must be provided.
All other arguments are optional, and depend on the XPL program being run.
.PP
.B xplsm
takes the following arguments:
.TP
.IR \-p ,\  \-\-program \  load\-file
This parameter specifies the XPL load file containing the binary image of
an XPL program.
It is executed under control of
.BR xplsm .
.TP
.IR \-f ,\  \-\-file \  blocksize
The random access files referenced from an XPL program as
.BR FILE(1) ,
.BR FILE(2) ,
etc. are to use
.I blocksize
for a record length.
Beware that the XPL program must know this size.
The default is 512.
.TP
.IR \-l ,\  \-\-lines\-per\-page \  lines
The XPL output files
.B OUTPUT(0)
and
.B OUTPUT(1)
are actually the same file, and
are intended to be used for printed output.
A form feed character (^L) is inserted automatically after
.I lines
lines have been written to
.BR OUTPUT(0) .
If
.I lines
is zero, then no form feed characters are automatically inserted.
The default is 60 lines.
The first character of a line written to
.B OUTPUT(1)
is examined for an ANSI carriage control character:
.PP
.RS
space	\- single space
.br
\'0'	\- double space
.br
\'\-'	\- triple space
.br
\'1'	\- new page; a form feed is inserted
.RE
.IP ""
The current number of lines is adjusted accordingly.
.TP
.IR \-m ,\  \-\-maxcore \  size
The size of the dynamic string area is set to
.I size.
It defaults to 65K.
.TP
.I \-\-trace
The XPL program begins execution as if it had issued a
.B CALL TRACE
before the first instruction is executed.
.TP
.I \-\-dump
If the XPL program issues a
.BR CALL\ EXIT ,
or dies due to a UNIX signal, a diagnostic
dump is written to stderr.
.TP
.I \-\-mmap
If
.B xplsm
was compiled with support for
.IR mmap (2),
then mmap is used to map the code section of the load file.
.TP
.I \-\-input filename
.nr dp \n(PD
.nr PD 0
.TP
.I \-\-input2 filename
.TP
.I \-\-input3 filename
.br
.nr PD \n(dp
The input files referred to as
.BR INPUT(0) ,
.BR INPUT(1) ,
etc. by the XPL program
are associated with
.IR filename .
The
.I filename
.RB ' - '
or
.B stdin
represents stdin.
If the first character of
.I filename
is a
.RB ' & ',
then the file is considered temporary, and it is unlinked after the initial
open.
If the first character of
.I filename
is a
.RB ' | ',
then the filename is interpreted as a command string, and a pipe is
opened to provide subsequent input lines to the XPL program.
Both
.B INPUT(0)
and
.B INPUT(1)
refer to the file specified by
.IR \-\-input ,
which defaults to stdin.
.TP
.I \-\-output filename
.nr PD 0
.TP
.I \-\-output2 filename
.TP
.I \-\-output3 filename
.TP
.I \-\-output4 filename
.br
.nr PD \n(dp
The output files referred to as
.BR OUTPUT(0) ,
.BR OUTPUT(1) ,
etc. by the XPL program
are associated with
.IR filename .
The
.I filename
.RB ' - '
or
.B stdout
represents stdout.
The
.I filename
.B stderr
represents stderr.
If the first character of
.I filename
is a
.RB ' & ',
then the file is considered temporary, and it is unlinked after the initial
open.
If the first character of
.I filename
is a
.RB ' | ',
then the filename is interpreted as a command string, and a pipe is
opened to read the subsequent output from the XPL program.
Both
.B OUTPUT(0)
and
.B OUTPUT(1)
refer to the file specified by
.IR \-\-output ,
which defaults to stdout,
and
.B OUTPUT(2)
refers to the file specified by
.I \-\-output2
which defaults to stderr.
.TP
.I \-\-file1 filename
.nr PD 0
.TP
.I \-\-file2 filename
.TP
.I \-\-file3 filename
.TP
.I \-\-file4 filename
.br
.nr PD \n(dp
The random access files referred to as
.BR FILE(1) ,
.BR FILE(2) ,
etc. by the XPL program
are associated with
.IR filename .
If the first character of
.I filename
is a
.RB ' & ',
then the file is considered temporary, and it is unlinked after the initial
open.
.TP
.I \-\-help
Prints a usage summary for
.B xplsm
on stderr.
.TP
.I \-\-version
Prints the current version for
.BR xplsm.
.TP
.IR parms ...
All remaining arguments are combined into a single XPL
string, each item separated from the next with a comma.
This string is available to the XPL program as
.BR PARMSTRING .
.SH "ABNORMAL TERMINATION"
Various errors can cause abnormal termination of the XPL program.
Usually, these errors are self-explanatory.
The error message "Abnormal termination of XPL program
.I name
.RI ( code )
at offset
.IR value "
indicates a
.I code
number that can further identify the cause of the error.
The offset
.I value
is the offset within the XPL program of the statement which
triggered the error.
The possible codes are:
.TP
500
The XPL program called the submonitor with an illegal service code.
.TP
.RI 60 n
The XPL program specified an invalid input file
.IR n .
.TP
.RI 70 n
The XPL program specified an invalid output file
.IR n .
.TP
.RI 80 n
The XPL program specified an invalid random access file
.IR n .
.TP
.RI 1 mnn
The XPL program encountered an I/O error on input file
.I m
with
.I errno
=
.IR nn .
.TP
.RI 2 mnn
The XPL program encountered an I/O error on output file
.I m
with
.I errno
=
.IR nn .
.TP
.RI 3 mnn
The XPL program encountered an I/O error on random access file
.I m
with
.I errno
=
.IR nn .
.TP
4000
The XPL program called EXIT.
.SH "SEE ALSO"
.IR xpl (1),
.IR errno (2)
.SH AUTHOR
David E. Bodenstab (imdave@mcs.net)
