/*
 * XPLSM    SUBMONITOR FOR THE XPL COMPILER GENERATOR SYSTEM
 *
 *                               DAVID B. WORTMAN
 *                               UNIVERSITY OF TORONTO
 *                               JANUARY 1972
 *
 *
 * Implementation in C
 *
 * Copyright 1984, 1996, 1997, David E. Bodenstab
 * All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include "s370.h"

#define	DIM(x)	(sizeof(x)/sizeof(x[0]))

typedef enum{ FALSE=0, TRUE=1 } boolean;

char stderrbuf[BUFSIZ];


/*
 * FILEBYTS DETERMINES THE BLOCKSIZE FOR DIRECT ACCESS FILE
 * I/O.  IT SHOULD BE EQUAL TO THE VALUE OF THE LITERAL
 * CONSTANT 'DISKBYTES' IN THE XCOM COMPILER FOR COMPILATION
 * TO WORK SUCCESSFULLY.  THE VALUE OF FILEBYTS WHICH IS
 * ASSEMBLED IN MAY BE OVERIDDEN BY THE 'FILE=NNNN' PARAMETER
 * ON THE OS/360 EXEC CARD.
 */

#define	FILEBYTS	512


/*
 * OS/370 <==> UNIX simulation structures
 */

enum  recfm { NA=0,	/* ..X means xlate between ASCII & EBCDIC */
	      F,   V,
	      FA,  VA,
	      FX,  VX,
	      FAX, VAX,
	    };
enum  dsorg { QSAM, BSAM };

struct  JFCB
  {
    const char * const	ddname;	/* DDNAME */
    const char		*fname;	/* UNIX filename */
    enum recfm		recfm;	/* optional recfm */
    int			lrecl;	/* optional lrecl */
  };

struct  IHADCB
  {
    /*
     * OS/370
     */
    const enum dsorg	dsorg;
    struct JFCB * const	jfcb;
    const int		macrf;		/* valid open(2) flags */
    enum recfm		recfm;
    int			lrecl;
    void		(**exlst)(struct IHADCB*);
    boolean * const	eodad;
    void		(*synad)(struct IHADCB*,int);
    boolean		flag;		/* TRUE if OPENed */
    void		(*in)();	/* get/read routine */
    void		(*out)();	/* put/write routine */
    /*
     * UNIX
     */
    struct
      {					/* save original values */
	enum recfm	recfm;
	int		lrecl;
      }
	save;
    const char		*fname;
    int			oflag;		/* open(2) flags */
    union
      {
	struct
	  {
	    FILE	*F;
	    boolean	firstline;	/* TRUE for the first line of the file */
	    int		LineNumber;	/* current line number */
	    int		Lcolpos;
	    int		Nspace;
	  }
	    qsam;
	struct
	  {
	    int		fd;
	    void	(*point)();	/* point routine */
	  }
	    bsam;
      }
	access;
  };

/*
 * ROUTINES TO PROVIDE DEFAULT DATASET INFORMATION IF NONE
 * IS PROVIDED BY JCL OR VOLUME LABELS.  IN PARTICULAR,
 * BLKSIZE, LRECL, BUFNO, AND RECFM INFORMATION.
 */

/*
 * ARRAY OF DEFAULT ATTRIBUTES USED BY GENXT
 */
struct DEFAULTS
  {
  int		lrecl;	/* DEFAULT LRECL */
  enum recfm	recfm;	/* DEFAULT RECFM */
  };

/*
 * DEFINE ATTRIBUTES PROVIDED FOR THE VARIOUS FILES
 *
 * INPUT(0), INPUT(1), OUTPUT(2)
 */
struct DEFAULTS INDFLT0 ={ 80, FX };

/*
 * OUTPUT(0), OUTPUT(1)
 */
struct DEFAULTS OUTDFLT0 ={ 133, FAX };

/*
 * INPUT(2), INPUT(3), OUTPUT(3)
 */
struct DEFAULTS INDFLT2 ={ 80, FX };

/*
 * DCB EXIT ROUTINE ENTRY POINTS
 */

void GENXT( struct IHADCB *, struct DEFAULTS * );

 void
INXT0( struct IHADCB *dcbp )
{
  GENXT( dcbp, &INDFLT0 );
}

 void
INXT2( struct IHADCB *dcbp )
{
  GENXT( dcbp, &INDFLT2 );
}

 void
OUTXT0( struct IHADCB *dcbp )
{
  GENXT( dcbp, &OUTDFLT0 );
}

/*
 * DCB EXIT LIST PROCESSING ROUTINE FOR OPEN EXITS
 */

 void
GENXT( struct IHADCB *dcbp, struct DEFAULTS *dp )
{
  if ( dcbp->recfm == NA )
    dcbp->recfm = dp->recfm;
  if ( dcbp->lrecl == 0 )
    dcbp->lrecl = dp->lrecl;
}

/*
 * EXIT LISTS FOR DCBS
 */

void (*INEXIT0)(struct IHADCB*) = INXT0;	/* INPUT0 */
void (*OUTXIT0)(struct IHADCB*) = OUTXT0;	/* OUTPUT0 */
void (*INEXIT2)(struct IHADCB*) = INXT2;	/* INPUT2 */
void (*OUTXIT2)(struct IHADCB*) = INXT0;	/* OUTPUT2 */

#define	INEXIT3	INEXIT2
#define	OUTXIT3	INEXIT2
#define	OUTXIT4	INEXIT2


/*
 * DEVICE CONTROL BLOCKS FOR THE SUBMONITOR
 */

extern struct JFCB program[], input[], output[], file[];

void ERRPGM( struct IHADCB*, int );
void INSYNAD( struct IHADCB*, int );
void OUTSYNAD( struct IHADCB*, int );
void FILESYND( struct IHADCB*, int );

boolean EODPGM = FALSE;

struct  IHADCB  PROGRAM ={ BSAM, program, O_RDONLY, NA, 0, NULL, &EODPGM, ERRPGM };

boolean INEOD = FALSE;

struct  IHADCB  INPUT0 ={ QSAM, input+0, O_RDONLY, NA, 0, &INEXIT0, &INEOD, INSYNAD };

#define	INPUT1	INPUT0		/* INPUT(0) & INPUT(1) ARE BOTH SYSIN */

struct  IHADCB  INPUT2 ={ QSAM, input+1, O_RDONLY, NA, 0, &INEXIT2, &INEOD, INSYNAD };
struct  IHADCB  INPUT3 ={ QSAM, input+2, O_RDONLY, NA, 0, &INEXIT3, &INEOD, INSYNAD };

struct  IHADCB  OUTPUT0 ={ QSAM, output+0, O_WRONLY, NA, 0, &OUTXIT0, NULL, OUTSYNAD };

#define	OUTPUT1	OUTPUT0		/* OUTPUT(0), OUTPUT(1) BOTH SYSPRINT */

struct  IHADCB  OUTPUT2 ={ QSAM, output+1, O_WRONLY, NA, 0, &OUTXIT2, NULL, OUTSYNAD };
struct  IHADCB  OUTPUT3 ={ QSAM, output+2, O_WRONLY, NA, 0, &OUTXIT3, NULL, OUTSYNAD };
struct  IHADCB  OUTPUT4 ={ QSAM, output+3, O_WRONLY, NA, 0, &OUTXIT4, NULL, OUTSYNAD };

/*
 * DCBS FOR THE DIRECT ACCESS FILES
 */

boolean FILEEOD = FALSE;

struct  IHADCB  FILE1 ={ BSAM, file+0, O_RDWR, F, FILEBYTS, NULL, &FILEEOD, FILESYND };
struct  IHADCB  FILE2 ={ BSAM, file+1, O_RDWR, F, FILEBYTS, NULL, &FILEEOD, FILESYND };
struct  IHADCB  FILE3 ={ BSAM, file+2, O_RDWR, F, FILEBYTS, NULL, &FILEEOD, FILESYND };
struct  IHADCB  FILE4 ={ BSAM, file+3, O_RDWR, F, FILEBYTS, NULL, &FILEEOD, FILESYND };


/*
 * DCB ADDRESS TABLE FOR ALL I/O ROUTINES
 *
 * THE FOUR SETS OF DCB ADDRESSES HEADED BY  'GETDCBS',
 * 'PUTDCBS', 'ARWDCBS', AND 'PGMDCB' MUST BE CONTIGUOUS
 * AND END WITH 'PGMDCB'.  THESE LISTS ARE USED AT JOB END
 * TO CLOSE ALL FILES BEFORE RETURNING TO OS
 *
 * DCB ADDRESSES FOR INPUT FILES:
 */

struct IHADCB *GETDCBS[] ={ &INPUT0, &INPUT1, &INPUT2, &INPUT3, };

#define	INPUTS	DIM(GETDCBS)-1	/* NUMBER OF INPUT FILES */
				/*    INPUT(I),  I = 0,1,2,3 */

/*
 * DCB ADDRESSES FOR OUTPUT FILES
 */

struct IHADCB *PUTDCBS[] ={ &OUTPUT0, &OUTPUT1, &OUTPUT2, &OUTPUT3, &OUTPUT4, };

#define	OUTPUTS	DIM(PUTDCBS)-1	/* NUMBER OF OUTPUT FILES */
				/*    OUTPUT(I),  I = 0,1,2,3,4 */

/*
 * DCB ADDRESS FOR DIRECT ACCESS FILES
 */

struct IHADCB *ARWDCBS[] ={ &FILE1, &FILE2, &FILE3, &FILE4, };

#define	FILES	DIM(ARWDCBS)	/* NUMBER OF DIRECT ACCESS FILES */
				/*    FILE (I,*),  I = 1,2,3,4 */

struct IHADCB *PGMDCB[] ={ &PROGRAM, NULL };

/*
 * UNIX SIMULATION OF DDNAMES:  EACH ARRAY CORRESPONDS
 * TO THE NAMED DDNAME SUFFIXED WITH AN INTEGER SUBSCRIPT
 */

struct JFCB program[1] ={ {"PROGRAM"} };
struct JFCB input[INPUTS] ={ {"SYSIN"}, {"INPUT2"}, {"INPUT3"} };
struct JFCB output[OUTPUTS] ={ {"SYSPRINT"}, {"SYSPUNCH"}, {"OUTPUT3"}, {"OUTPUT4"} };
struct JFCB file[FILES] ={ {"FILE1"}, {"FILE2"}, {"FILE3"}, {"FILE4"} };

/*
 * DEFINE THE REGISTERS USED TO PASS PARAMETERS TO THE
 * SUBMONITOR FROM THE XPL PROGRAM
 */

#define	SVCODE	1	/* CODE INDICATING SERVICE REQUESTED */
#define	PARM1	0	/* FIRST PARAMETER */
#define	PARM2	2	/* SECOND PARAMETER */

/*
 * DEFINE THE SERVICE CODES USED BY THE XPL PROGRAM TO
 * INDICATE SERVICE REQUESTS TO THE SUBMONITOR
 */

#define	GETC	4	/* SEQUENTIAL INPUT */
#define	PUTC	8	/* SEQUENTIAL OUTPUT */
#define	TRC	12	/* INITIATE TRACING */
#define	UNTR	16	/* TERMINATE TRACING */
#define	EXDMP	20	/* FORCE A CORE DUMP */
#define	GTIME	24	/* RETURN TIME AND DATE */
#define	RSVD1	28	/* (UNUSED) */
#define	CLTRP	32	/* CLOCK_TRAP */
#define	INTRP	36	/* INTERRUPT_TRAP */
#define	MONTR	40	/* MONITOR */
#define	RSVD5	44	/* (UNUSED) */
#define	RSVD6	48	/* (UNUSED) */


/*
 * GENERATE THE SERVICE CODES FOR DIRECT ACCESS FILE I/O
 * BASED ON THE NUMBER OF FILES AVAILABLE  (&FILES)
 */

#define	FILEORG	(RSVD6+4)		/* ORIGIN FOR THE FILE SERVICE CODES */

#define	RD1	FILEORG+8*(1-1)		/* CODE FOR READING FILE1 */
#define	WRT1	FILEORG+8*(1-1)+4	/* CODE FOR WRITING FILE1 */
#define	RD2	FILEORG+8*(2-1)		/* CODE FOR READING FILE2 */
#define	WRT2	FILEORG+8*(2-1)+4	/* CODE FOR WRITING FILE2 */
#define	RD3	FILEORG+8*(3-1)		/* CODE FOR READING FILE3 */
#define	WRT3	FILEORG+8*(3-1)+4	/* CODE FOR WRITING FILE3 */
#define	RD4	FILEORG+8*(4-1)		/* CODE FOR READING FILE4 */
#define	WRT4	FILEORG+8*(4-1)+4	/* CODE FOR WRITING FILE4 */

/*
 * DEFINE REGISTER USAGE
 */

#define	RTN	3	/* REGISTER CONTAINING COMPLETION CODE RETURNED BY THE PROGRAM */
#define	CBR	12	/* LINKAGE REGISTER USED FOR CALLS TO THE SUBMONITOR */
#define	SELF	15	/* REGISTER THAT ALWAYS CONTAINS THE ADDRESS OF THE SUBMONITOR ENTRY POINT */

/*
 * FLAG BITS USED TO CONTROL SUBMONITOR OPERATION
 */

#define	DUMPBIT	0x08	/* GIVE A CORE DUMP FOR I/O ERRORS */

/*
 * DEFINE ABEND CODES ISSUED BY THE SUBMONITOR
 */

#define	OPENABE	100	/* UNABLE TO OPEN ONE OF THE FILES: PROGRAM, SYSIN, OR SYSPRINT */
#define	PGMEOD	200	/* UNEXPECTED END OF FILE WHILE READING IN THE XPL PROGRAM */
#define	PGMERR	300	/* SYNAD ERROR WHILE READING IN THE XPL PROGRAM */
#define	COREABE	400	/* XPL PROGRAM WON'T FIT IN THE AMOUNT OF MEMORY AVAILABLE */
#define	CODEABE	500	/* INVALID SERVICE CODE FROM THE XPL PROGRAM */
#define	PGMABE	600	/* BYTSBLK IS NOT EQUAL TO //PROGRAM LRECL */
#define	PFADDR	700	/* ADDRESSING ERROR FOR OUTPUT ADDRESS */
#define	OUTSYND	800	/* SYNAD ERROR ON OUTPUT FILE */
#define	PFABE	900	/* INVALID OUTPUT FILE SPECIFIED */
#define	INSYND	1000	/* SYNAD ERROR ON INPUT FILE */
#define	INEODAB	1200	/* END OF FILE ERROR ON INPUT FILE */
#define	GFABE	1400	/* INVALID INPUT FILE SPECIFIED */
#define	GFADDR	1600	/* ADDRESSING ERROR FOR INPUT ADDRESS */
#define	FLSYND	2000	/* SYNAD ERROR ON DIRECT ACCESS FILE */
#define	FLEOD	2200	/* END OF FILE ERROR ON DIRECT ACCESS FILE */
#define	FLADDR	2400	/* ADDRESSING ERROR FOR FILE ADDRESS */
#define	USERABE	4000	/* XPL PROGRAM CALLED EXIT TO FORCE A CORE DUMP */

/*
 * DSECT WHICH DEFINES THE FORMAT OF BINARY PROGRAM CONTROL
 * INFORMATION AND THE STARTING POINT FOR PROGRAMS
 */

struct  FILECTRL
  {
    int BYTSCODE;	/* NUMBER OF BYTES OF CODE */
    int BYTSDATA;	/* NUMBER OF BYTES OF DATA AREA */
    int BLKSCODE;	/* NUMBER OF RECORDS OF CODE */
    int BLKSDATA;	/* NUMBER OF RECORDS OF DATA AREA */
    int BYTSBLK;	/* BLOCKSIZE OF THE XPL PROGRAM FILE */
    int BYTSFULL;	/* NUMBER OF BYTES OF CODE ACTUALLY USED IN THE LAST RECORD OF CODE */
    int DATABYTS;	/* NUMBER OF BYTES OF DATA ACTUALLY USED IN THE LAST RECORD OF DATA */
    int pad[8];		/* REMAINDER OF THE CONTROL BLOCK IS UNUSED */
  };

#define	CODEBEGN sizeof(struct FILECTRL)	/* FIRST EXECUTABLE INSTRUCTION IN THE XPL PROGRAM */

/*
 * CONSTANTS USED DURING INITIALIZATION
 */

char *ACORE;	/* ADDRESS OF CORE FOR THE PROGRAM */
long CORESIZE;	/* SIZE OF CORE IN BYTES */
char *CONTROL;	/* ADDRESS OF PARAMETER STRING PASSED TO THE SUBMONITOR BY OS/360 */

struct
  {			/*  CORE REQUEST CONTROL BLOCK */
    long COREMIN;	/* MINIMUM AMOUNT OF CORE REQUESTED */
    long COREMAX;	/* MAXIMUM AMOUNT OF CORE REQUESTED */
  } COREREQ = { 30000, 300000 };

long FREEUP = 2*FILEBYTS;	/* AMOUNT OF CORE TO RETURN TO OS */
long ALTFREE = 4*FILEBYTS;	/* AMOUNT OF CORE FREED FOR ALTER */

/*
 * If EBCDIC is TRUE, then recfm's ..X will result in normal UNIX files
 *  being translated to/from EBCDIC on read/write.  Recfm=F..X will cause
 *  records to be padded to LRECL for input.  Output will always be UNIX
 *  style lines terminated with a NL.
 *
 *  Recfm's FA* will treat normal UNIX files as IBM style fixed length
 *  records, and no ASCII to/from EBCDIC translation will occur.
 *
 *  Recfm VA* is not supported.
 *
 * If EBCDIC is FALSE, then no translation between ASCII/EBCDIC is done.
 *  All files are normal UNIX files.  Recfm=FA*X* will pad records to
 *  LRECL on input.  Output will always be UNIX style lines terminated
 *  with NL.  Recfm's [FV]A* are equivalent to [FV]A*X.
 */
boolean EBCDIC = FALSE;

/*
 * BLOCK OF PARAMETERS PASSED TO THE XPL PROGRAM
 */

struct
  {
    int   PGMPARMS;	/* R0  DESCRIPTOR FOR OS/360 PARAMETER STRING */
    char  *CORETOP;	/* R1  ADDRESS OF TOP OF CORE */
    char  *CODEADR;	/* R2  ADDRESS OF START OF 1ST RECORD OF THE XPL PROGRAM */
    char  *DATADR;	/* R3  ADDRESS OF THE START OF THE XPL PROGRAM'S DATA AREA */
    int R4_R15[12];	/* R4 - R15 */
  } s370regs = { 0,0,0,0, };

struct  psw psw;

/*
 * DATA AREA FOR THE SUBMONITOR
 */

int RTNSV;	/* SAVE COMPLETION CODE RETURNED BY THE XPL PROGRAM */
char FLAGS;	/* SUBMONITOR CONTROL FLAGS */


/*
 * SUBMONITOR  INITIALIZATION
 */

void		abend( int );
boolean		ddname();
int		fetchw( void * );
char		*ucase();
boolean		s370( struct psw *, int * );
void		s370init( unsigned char *, int, struct psw *, int * );
void		ABEND( int );
void		CLOSE( struct IHADCB * );
void		ENTRY();
void		GET( unsigned char **, int *, unsigned char ** );
void		GETIME( int *, int * );
void		INERR( int, int );
void		OPEN( struct IHADCB*, int );
void		PUT( unsigned int *, int * );
void		READ( int, unsigned char **, unsigned int * );
void		READPGM( char* );
void		USEREXIT();
void		WRITE( int, unsigned char **, unsigned int * );

/*
 * TRUE if tracing
 */
extern boolean	T;


 int
main( argc, argv )
  char **argv;
{
  char *p, *org;
  int c, i, l;
  struct FILECTRL *fcp;
  boolean error = FALSE;

  while( (c=getopt(argc,argv,"p:i:o:f:")) != EOF )
    switch( c )
      {
    case 'p':
      if ( !ddname(optarg,program,DIM(program)) )
	{
	  fprintf( stderr, "//PROGRAM DD: specification error\n" );
	  error = TRUE;
	}
      break;
    case 'i':
      if ( !ddname(optarg,input,DIM(input)) )
	{
	  fprintf( stderr, "//INPUT DD: specification error\n" );
	  error = TRUE;
	}
      break;
    case 'o':
      if ( !ddname(optarg,output,DIM(output)) )
	{
	  fprintf( stderr, "//OUTPUT DD: specification error\n" );
	  error = TRUE;
	}
      break;
    case 'f':
      if ( !ddname(optarg,file,DIM(file)) )
	{
	  fprintf( stderr, "//FILE DD: specification error\n" );
	  error = TRUE;
	}
      break;
    case '?':
      error = TRUE;
      break;
      }

  if ( error )
    exit( 2 );

  /* INPUT0 & INPUT1 default to stdin */
  if ( ! input[0].fname )
    input[0].fname = "-";
  /* OUTPUT0 & OUTPUT1 default to stdout */
  if ( ! output[0].fname )
    output[0].fname = "-";
  /* OUTPUT2 defaults to stdout */
  if ( ! output[1].fname )
    output[1].fname = "-";

  for( l=0,i=optind; i<argc; i++ )
    l += strlen(argv[i]) + 1;
  if ( l > 0 )
    {
      strcpy( p = malloc(l), argv[optind++] );
      for( ; optind<argc; ++optind )
	strcat( strcat(p,","), argv[optind] );
    }
  else
    p = NULL;

   CONTROL = p;	/* SAVE ADDRESS FOR THE TRACE ROUTINE */

   FLAGS = 0;	/* RESET ALL FLAGS */

/*
 * CREATE AN XPL STRING DESCRIPTOR FOR THE OS/360 PARAMETER
 * STRING SO IT CAN BE PASSED TO THE XPL PROGRAM
 */

  if ( CONTROL != NULL )
    for( p=CONTROL; *p; ++p )
      {
	if ( strncmp(p,"ALTER",5) == 0 )	/* check for ALTER */
	  {
	    FREEUP = ALTFREE;
	    p += 4;
	  }
	if ( strncmp(p,"TRACE",5) == 0 )	/* check for TRACE */
	  {
	    T = TRUE;
	    setvbuf(stderr, stderrbuf, _IOFBF, sizeof(stderrbuf));
	    p += 4;
	  }
	if ( strncmp(p,"MIN=",4) == 0 )		/* check for MIN=NNN */
	  {
	    COREREQ.COREMIN = atoi(p+4);
	    p += 4;
	  }
	if ( strncmp(p,"MAX=",4) == 0 )		/* check for MAX=NNN */
	  {
	    COREREQ.COREMAX = atoi(p+4);
	    p += 4;
	  }
	if ( strncmp(p,"FREE=",5) == 0 )	/* check for FREE=NNN */
	  {
	    FREEUP = atoi(p+5);
	    p += 5;
	  }
	if ( strncmp(p,"DUMP",4) == 0 )		/* check for DUMP */
	  {
	    FLAGS |= DUMPBIT;
	    p += 3;
	  }
	if ( strncmp(p,"ASCII",5) == 0 )	/* check for ASCII */
	  {
	    EBCDIC = FALSE;
	    p += 5;
	  }
	if ( strncmp(p,"EBCDIC",6) == 0 )	/* check for EBCDIC */
	  {
	    EBCDIC = TRUE;
	    p += 6;
	  }
	if ( strncmp(p,"FILE=",5) == 0 )	/* check for FILE=NNN */
	  {
	    struct IHADCB **d;
	    l = atoi(p+5);
	    for( d=ARWDCBS; d<ARWDCBS+FILES; ++d )
	      (*d)->lrecl = l;	/* SET NEW RECORD LENGTH */
	    ALTFREE = l*4;
	    p += 5;
	  }
      }
    
/*
 * OPEN THE BINARY PROGRAM FILE
 */

  OPEN( &PROGRAM, O_RDONLY );

  if ( ! PROGRAM.flag )
    ABEND( OPENABE );

/*
 * NOW OBTAIN SPACE IN MEMORY FOR THE XPL PROGRAM AND ITS
 * FREE STRING AREA.  A GETMAIN IS ISSUED TO OBTAIN AS MUCH
 * MEMORY AS POSSIBLE WITHIN THE PARTITION.  THEN A FREEMAIN
 * IS ISSUED TO RETURN THE AMOUNT OF MEMORY SPECIFIED BY
 * THE VARIABLE 'FREEUP' TO OS/360 FOR USE AS WORK SPACE.
 * OS/360 NEEDS SPACE FOR FOR DYNAMICALLY CREATING BUFFERS
 * FOR THE SEQUENTIAL INPUT AND OUTPUT FILES AND FOR
 * OVERLAYING I/O ROUTINES.
 *
 * THE AMOUNT OF CORE REQUESTED FROM OS/360 CAN BE ALTERED
 * WITH THE 'MAX=NNNN' AND 'MIN=MMMM' PARAMETERS TO THE
 * SUBMONITOR.  THE AMOUNT OF CORE RETURNED TO OS/360 CAN BE
 * ALTERED WITH THE 'FREE=NNNN' OR THE 'ALTER' PARAMETER
 *
 * MEMORY REQUEST IS DEFINED BY THE CONTROL BLOCK STARTING AT
 * 'COREREQ'.  THE DESCRIPTION OF THE MEMORY SPACE OBTAINED
 * IS PUT INTO THE CONTROL BLOCK STARTING AT 'ACORE'.
 */

  FREEUP &= 0xFFFFFFF8;	/* INSURE DOUBLEWORD ALIGNMENT */
  COREREQ.COREMIN &= 0xFFFFFFF8;
  COREREQ.COREMAX &= 0xFFFFFFF8;

  if ( (int)(ACORE=sbrk(COREREQ.COREMAX)) == -1 )
    {
      struct rlimit rlimit;
      char *now, *top;

      now=sbrk(0);
      getrlimit(RLIMIT_DATA,&rlimit);
      top=now+rlimit.rlim_max;

      if ( top-now >= COREREQ.COREMIN )
	ACORE = sbrk( CORESIZE=top-now );
      else
	abend( 0x0080A000 );
    }
  else
    CORESIZE = COREREQ.COREMAX;

  sbrk( -FREEUP );
  s370regs.CORETOP = sbrk(0);	/* ADDRESS OF TOP OF USABLE CORE (BECOMES 'FREELIMIT') */

  CORESIZE -= FREEUP;

  s370init( org = ACORE, CORESIZE, &psw, (int*) &s370regs );

  /* Relocate CONTROL string at base of memory */
  if ( CONTROL == NULL )
    s370regs.PGMPARMS = NULL;
  else
    {
      unsigned char* tr_ae( unsigned char*, int );

      org += (l = strlen( CONTROL )) + 7;
      org = (char*)((int)org & 0xfffffff8);

      if ( EBCDIC )
	memcpy( ACORE, tr_ae(CONTROL,l), l );
      else
	memcpy( ACORE, CONTROL, l );

      s370regs.PGMPARMS = ((l-1)<<24) | (int)ACORE;
    }

/*
 * READ IN THE BINARY XPL PROGRAM AS SPECIFIED BY THE
 *
 * //PROGRAM  DD  .....
 *
 * CARD.  IT IS ASSUMED THAT THE BINARY PROGRAM IS IN STANDARD
 * XPL SYSTEM FORMAT.
 *
 * THE 1ST RECORD OF THE BINARY PROGRAM FILE SHOULD BEGIN
 * WITH A BLOCK OF INFORMATION DESCRIBING THE CONTENTS OF
 * THE FILE.  THE FORMAT OF THIS BLOCK IS GIVEN IN THE DSECT
 * 'FILECTRL' AT THE END OF THIS ASSEMBLY.
 */

  s370regs.CODEADR = org;

  p = (char*)(fcp = (struct FILECTRL *)org);

  READPGM( p );

  if ( fetchw(&fcp->BYTSCODE) -
      fetchw(&fcp->BYTSBLK) +
      fetchw(&fcp->BYTSFULL) +
      fetchw(&fcp->BYTSDATA) > CORESIZE )
    ABEND( COREABE );

  if ( PROGRAM.lrecl != fetchw(&fcp->BYTSBLK) )
    ABEND( PGMABE );

  for( c=fetchw(&fcp->BLKSCODE); c>1; --c )
    {
      p += fetchw( &fcp->BYTSBLK );
      READPGM( p );
    }

  p += fetchw( &fcp->BYTSFULL );

  p += 3;
  p = (char*)((int)p &0xFFFFFFFC);

  s370regs.DATADR = p;

  for( c=fetchw(&fcp->BLKSDATA); c>0; --c )
    {
      READPGM( p );
      p += fetchw( &fcp->BYTSBLK );
    }

   CLOSE( &PROGRAM );

/*
 * ZERO CORE BETWEEN FREEBASE AND FREELIMIT TO REDUCE
 * PRINTING IN THE EVENT OF ABEND DUMPS
 */

  memset( p, 0, s370regs.CORETOP-p );

/*
 * CODE TO BRANCH TO THE XPL PROGRAM
 */

  ( (int*) &s370regs ) [SELF] = (int) ENTRY;
  ( (int*) &s370regs ) [CBR] = 0;

  psw.ia = (unsigned char*)org + CODEBEGN;

  while( s370(&psw,(int*)&s370regs) && psw.ia != 0 )
    {
      if ( psw.ia == (unsigned char*)ENTRY )
	{
	  ENTRY( &s370regs );
	  psw.ia = (unsigned char*)(((int*)&s370regs)[CBR]);
	}
    }

/*
 * THE XPL PROGRAM RETURNS HERE AT THE END OF EXECUTION
 */

  RTNSV = ((int*)&s370regs)[RTN];

  ((int*)&s370regs)[1] = UNTR;
  ENTRY( &s370regs );

  CLOSE( &INPUT0 );
  CLOSE( &OUTPUT0 );

    {
      struct IHADCB **dcbp;

      for( dcbp=GETDCBS; *dcbp; ++dcbp )
	CLOSE( *dcbp );
    }

  return( RTNSV );
}


 void
READPGM( p )
  char *p;
{
  (*PROGRAM.in)( &PROGRAM, p );

  if ( EODPGM )
    ABEND( PGMEOD );
}


/*
 * INPUT - OUTPUT  ERROR ROUTINES
 *
 * SYNAD AND EOD ERROR ROUTINES FOR INITIAL LOADING OF THE
 * XPL PROGRAM
 */

 void
ERRPGM( dcbp, err )
  struct IHADCB *dcbp;
{
  const struct JFCB *ddp = dcbp->jfcb;

  if ( err == 0 )
    fprintf( stderr, "I/O error on //%s DD DSNAME=\"%s\"\n", ddp->ddname, ddp->fname );
  else
    perror( "PROGRAM" );
  ABEND( PGMERR );
}


/*
 * SYNAD AND EOD ROUTINES FOR INPUT(I),  I = 0,1, ...  ,&INPUTS
 */


 void
INSYNAD( dcbp, err )
  struct IHADCB *dcbp;
{
  int i;
  const struct JFCB *ddp = dcbp->jfcb;

  i = 0;
  while( GETDCBS[i] != dcbp )
    ++i;

  if ( err == 0 )
    fprintf( stderr, "I/O error on //%s DD DSNAME=\"%s\"\n", ddp->ddname, ddp->fname );
  else
    perror( dcbp->fname );
  INERR( INSYND, i );
}


 void
INERR( code, n )
{
  ABEND( code + n );
}


/*
 * SYNAD ERROR ROUTINES FOR OUTPUT FILES
 */

 void
OUTSYNAD( dcbp, err )
  struct IHADCB *dcbp;
{
  int i;
  const struct JFCB *ddp = dcbp->jfcb;

  i = 0;
  while( PUTDCBS[i] != dcbp )
    ++i;

  if ( ! dcbp->flag )
    fprintf( stderr, "//%s DD missing\n", ddp->ddname );
  else
    if ( err == 0 )
      fprintf( stderr, "I/O error on //%s DD DSNAME=\"%s\"\n", ddp->ddname, ddp->fname );
    else
      perror( dcbp->fname );
  ABEND( OUTSYND+i );
}

/*
 * SYNAD AND EOD ROUTINES FOR DIRECT ACCESS FILE I/O
 */

 void
FILESYND( dcbp, err )
  struct IHADCB *dcbp;
{
  int i;
  const struct JFCB *ddp = dcbp->jfcb;

  i = 1;
  while( ARWDCBS[i] != dcbp )
    ++i;

  if ( ! dcbp->flag )
    fprintf( stderr, "//%s DD missing\n", ddp->ddname );
  else
    if ( err == 0 )
      fprintf( stderr, "I/O error on //%s DD DSNAME=\"%s\"\n",
	      ddp->ddname, ddp->fname );
    else
      perror( dcbp->fname );
  ABEND( FLSYND+i );
}


/*
 * ABEND ROUTINE FOR ALL I/O ERRORS
 *
 * ALL FILES WILL BE CLOSED UNLESS AN ABEND DUMP HAS BEEN
 * REQUESTED.
 */

 void
ABEND( code )
{
  register struct IHADCB **dcbp;

  if ( ! (FLAGS & DUMPBIT) )
    {
      CLOSE( &INPUT0 );
      CLOSE( &OUTPUT0 );

      for( dcbp=GETDCBS; *dcbp; ++dcbp )
	CLOSE( *dcbp );
    }

  abend( code );
}


/*
 * DISPATCHER FOR ALL SERVICE REQUESTS FROM THE XPL PROGRAM
 */

 void
ENTRY( regp )
  register int *regp;
{

  if ( T )
    fprintf( stderr, "ENTRY[%2d]  PARM1=%8.8X  PARM2=%8.8X\n", regp[SVCODE],
	    regp[PARM1], regp[PARM2] );

  switch( regp[SVCODE] )
    {
  default:
      ABEND( CODEABE );
      break;
  case TRC:
      if ( ! T )
	{
	T = TRUE;
	fflush( stderr );
	setvbuf(stderr, stderrbuf, _IOFBF, sizeof(stderrbuf));
	}
      break;
  case UNTR:
      if ( T )
	{
	T = FALSE;
	fflush( stderr );
	setbuf( stderr, NULL );
	}
      break;
  case GETC:
      GET( (unsigned char**) regp+PARM1, (int*) regp+PARM2, (unsigned char **) regp+SVCODE );
      break;
  case PUTC:
      PUT( regp+PARM1, regp+PARM2 );
      break;
  case EXDMP:
      USEREXIT();
      break;
  case GTIME:
      GETIME( regp+PARM1, regp+SVCODE );
      break;
  case RD1:
  case RD2:
  case RD3:
  case RD4:
      READ( (regp[SVCODE]-FILEORG)/8, (unsigned char **) regp+PARM1, regp+PARM2 );
      break;
  case WRT1:
  case WRT2:
  case WRT3:
  case WRT4:
      WRITE( (regp[SVCODE]-FILEORG)/8, (unsigned char **) regp+PARM1, regp+PARM2 );
      break;
  case CLTRP:
      break;
  case INTRP:
      break;
  case MONTR:
      break;
  case RSVD1:
  case RSVD5:
  case RSVD6:
      break;
    }

  if ( T )
    fprintf( stderr, "RETURN[%2d] PARM1=%8.8X  PARM2=%8.8X\n", regp[SVCODE],
	    regp[PARM1], regp[PARM2] );
}


/*
 * ROUTINE TO FORCE AN ABEND DUMP WHEN REQUESTED BY THE
 * XPL PROGRAM BY MEANS OF THE STATEMENT:
 *
 *     CALL  EXIT  ;
 */

 void
USEREXIT()
{
  FLAGS |= DUMPBIT;

  ABEND( USERABE );
}


/*
 * INPUT ROUTINE FOR READING SEQUENTIAL INPUT FILES
 *
 *
 * INPUT TO THIS ROUTINE IS:
 *
 *   PARM1   ADDRESS OF THE NEXT AVAILABLE SPACE IN THE PROGRAMS
 *           DYNAMIC STRING AREA  (FREEPOINT)
 *
 *   SVCODE  THE SERVICE CODE FOR INPUT
 *
 *   PARM2   A SUBCODE DENOTING WHICH INPUT FILE,
 *           INPUT(I),     I = 0,1, ... ,&INPUTS
 *
 * THE ROUTINE RETURNS:
 *
 *   PARM1   A STANDARD XPL STRING DESCRIPTOR POINTING AT THE INPUT
 *           RECORD WHICH IS NOW AT THE TOP OF THE STRING AREA
 *
 *   SVCODE  THE NEW VALUE FOR FREEPOINT, UPDATED BY THE LENGTH OF
 *           THE RECORD JUST READ IN
 *
 *
 * A STANDARD XPL STRING DESCRIPTOR HAS:
 *
 *     BITS  0-7                 (LENGTH - 1) OF THE STRING
 *     BITS  8-31                ABSOLUTE ADDRESS OF THE STRING
 */

 void
GET( string, n, freepoint )
  unsigned char **string;	/* pointer to top of XPL string area */
  register int *n;
  unsigned char **freepoint;	/* updated pointer to top of XPL string area */
{
  register struct IHADCB *dcbp;

  if ( *n<0 || *n>INPUTS )
    INERR( GFABE, *n );

  if ( ! (dcbp=GETDCBS[*n])->jfcb )
    INERR( INEODAB, *n );

  if ( ! dcbp->flag )
    {
      OPEN( dcbp, O_RDONLY );

      if ( ! dcbp->flag )
	{
	  *freepoint = *string;
	  *string = NULL;
	  return;
	}
    }

  if ( *string < (unsigned char*)ACORE || *string >= (unsigned char*)ACORE+CORESIZE )
    ABEND( GFADDR );

  (*(dcbp->in))( dcbp, *string );

  if ( *(dcbp->eodad) )
    {
      CLOSE( dcbp );

      *freepoint = *string;
      *string = NULL;
      return;
    }

  *freepoint = *string + strlen(*string);
  *((int*)string) = ((strlen(*string)-1)<<24) | ((unsigned int)(*string));
}


/*
 * ROUTINE FOR WRITING SEQUENTIAL OUTPUT FILES
 *
 *
 * INPUT TO THIS ROUTINE:
 *
 *   PARM1   XPL STRING DESCRIPTOR OF THE STRING TO BE OUTPUT
 *
 *   PARM2   SUBCODE INDICATING  OUTPUT(I),  I = 0,1, ... ,&OUTPUTS
 *
 *   SVCODE  THE SERVICE CODE FOR OUTPUT
 *
 *
 * THE STRING NAMED BY THE DESCRIPTOR IS PLACED IN THE NEXT
 * OUTPUT BUFFER OF THE SELECTED FILE.  IF THE STRING IS
 * SHORTER THAN THE RECORD LENGTH OF THE FILE THEN THE
 * REMAINDER OF THE RECORD IS PADDED WITH BLANKS.  IF THE
 * STRING IS LONGER THAN THE RECORD LENGTH OF THE FILE
 * THEN IT IS TRUNCATED ON THE RIGHT TO FIT.  IF THE SUBCODE
 * SPECIFIES OUTPUT(0) THEN A SINGLE BLANK IS CONCATENATED
 * ON TO THE FRONT OF THE STRING TO SERVE AS CARRIAGE CONTROL.
 */

short CURLIN;		/* CURRENT LINES FOR OUTPUT(0) & (1) */
short MAXLIN ={ 60 };	/* MAX LINES FOR OUTPUT(0) & (1) */

 void
PUT( string, n )
  unsigned int *string;	/* pointer to the XPL descriptor */
  register int *n;
{
  register unsigned char *p, *q;
  extern unsigned char AE[], EA[];
  unsigned char buffer[BUFSIZ+1];
  int l;
  boolean isX;
  struct IHADCB *dcbp;

  if ( *n<0 || *n>OUTPUTS )
    INERR( PFABE, *n );

  if ( ! (dcbp=PUTDCBS[*n])->flag )
    {
      OPEN( dcbp, O_WRONLY );

      if ( ! dcbp->flag )
	OUTSYNAD( dcbp, 0 );
    }

  isX = EBCDIC && (dcbp->recfm==FX || dcbp->recfm==VX || dcbp->recfm==FAX || dcbp->recfm==VAX);

  if ( *string == 0 )
    l = 0;
  else
    l = ((*string >> 24) & 0xFF) + 1;

  for( p=buffer, q=(unsigned char*)((*string)&0x00FFFFFF); l>0; --l )
    {
      if ( q < (unsigned char*)ACORE || q >= (unsigned char*)ACORE+CORESIZE )
	ABEND( PFADDR );
      if ( (*++p = *q++) == 0 )
	*p = isX? AE[' '] : ' ';
    }

  p[1] = 0;

  p = buffer + 1;

  if ( *n == 0 )	/* OUTPUT0 */
    {
      if ( CURLIN >= MAXLIN )
	{
	  *--p = isX? AE['1'] : '1';
	  CURLIN = 1;
	}
      else
	{
	  *--p = isX? AE[' '] : ' ';
	  ++CURLIN;
	}
    }
  else
    {
      if ( *n == 1 )	/* OUTPUT1 */
	{
	  if ( isX )
	    {
	      switch( EA[*p] )
		{
	      case '1': CURLIN = 1; break;
	      case '-': CURLIN += 3;  break;
	      case '0': CURLIN += 2;  break;
	      default:  CURLIN += 1;  break;
	      case '+':     break;
		}
	    }
	  else
	    {
	      switch( *p )
		{
	      case '1': CURLIN = 1; break;
	      case '-': CURLIN += 3;  break;
	      case '0': CURLIN += 2;  break;
	      default:  CURLIN += 1;  break;
	      case '+':     break;
		}
	    }
	}
    }

  (*(dcbp->out))( dcbp, p );
}


/*
 * READ ROUTINE FOR DIRECT ACCESS FILE I/O
 *
 * INPUT TO THIS ROUTINE IS:
 *
 *   PARM1   CORE ADDRESS TO READ THE RECORD INTO
 *
 *   SVCODE  SERVICE CODE INDICATING WHICH FILE TO USE
 *
 *   PARM2   RELATIVE RECORD NUMBER   0,1,2,3,...
 */

 void
READ( n, addr, record )
  unsigned char **addr;
  unsigned int *record;
{
  struct IHADCB *dcbp;

  dcbp = ARWDCBS[n];

  if ( ! dcbp->flag )
    {
      OPEN( dcbp, O_RDWR );

      if ( ! dcbp->flag )
	FILESYND( dcbp, 0 );
    }

  (*(dcbp->access.bsam.point))( dcbp, *record );

  if ( *addr < (unsigned char*)ACORE || *addr >= (unsigned char*)ACORE+CORESIZE )
    ABEND( FLADDR+n );

  (*(dcbp->in))( dcbp, *addr );

  if ( *(dcbp->eodad) )
    ABEND( FLEOD+n );
}

/*
 * WRITE ROUTINE FOR DIRECT ACCESS FILE I/O
 *
 * INPUT TO THIS ROUTINE IS:
 *
 *   PARM1   CORE ADDRESS TO READ THE RECORD FROM
 *
 *   SVCODE  SERVICE CODE INDICATING WHICH FILE TO USE
 *
 *   PARM2   RELATIVE RECORD NUMBER   0,1,2, ...
 */

 void
WRITE( n, addr, record )
  unsigned char **addr;
  unsigned int *record;
{
  struct IHADCB *dcbp;

  dcbp = ARWDCBS[n];

  if ( ! dcbp->flag )
    {
      OPEN( dcbp, O_RDWR );

      if ( ! dcbp->flag )
	FILESYND( dcbp, 0 );
    }

  (*(dcbp->access.bsam.point))( dcbp, *record );

  if ( *addr < (unsigned char*)ACORE || *addr >= (unsigned char*)ACORE+CORESIZE )
    ABEND( FLADDR+n );

  (*(dcbp->out))( dcbp, *addr );
}


/*
 * TIME AND DATE FUNCTIONS
 *
 * RETURNS TIME OF DAY IN HUNDREDTHS OF A SECOND IN REGISTER
 * PARM1  AND THE DATE IN THE FORM  YYDDD IN REGISTER SVCODE
 */

#include <time.h>

struct tm *localtime();

 void
GETIME( t, d )
  int *t, *d;
{
  long l;
  struct tm *p;

  time( &l );
  p = localtime( &l );

  *t = 100 * (60*60*(p->tm_hour) + 60*(p->tm_min) + p->tm_sec);

  *d = 1000 * (p->tm_year) + p->tm_yday + 1;
}


/*
 * I/O routines
 */

unsigned char AE[] ={
  0xFF,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,
  0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,
  0x40,0x5A,0x7F,0x7B,0x5B,0x6C,0x50,0x7D,0x4D,0x5D,0x5C,0x4E,0x6B,0x60,0x4B,0x61,
  0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0x7A,0x5E,0x4C,0x7E,0x6E,0x6F,
  0x7C,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,
  0xD7,0xD8,0xD9,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,0x40,0x40,0x40,0x40,0x6D,
  0x40,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x91,0x92,0x93,0x94,0x95,0x96,
  0x97,0x98,0x99,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0x40,0x4F,0x40,0x5F,0x40,
  };

unsigned char EA[] ={ 
  "................................................................"
  " ...........<(+|&.........!$*);~-/.........,%_>?..........:#@'=\""
  ".abcdefghi.......jklmnopqr........stuvwxyz......................"
  ".ABCDEFGHI.......JKLMNOPQR........STUVWXYZ......0123456789......"
  };


/*
 * Translate EBCDIC to ASCII; truncate any trailing spaces
 */
 char*
tr_ea( register unsigned char *p )
{
  static unsigned char buf[BUFSIZ+1];
  register unsigned char *q = buf, *nonblank = q, c;

  while( (c = *p++) )
    if ( (*q++ = EA[c]) != ' ' )
      nonblank = q;
  *nonblank = 0;

  return( (char*)buf );
}

/*
 * Translate ASCII to EBCDIC; pad with spaces to given length
 */
 unsigned char*
tr_ae( register unsigned char *p, int l )
{
  static unsigned char buf[BUFSIZ+1];
  register unsigned char *q = buf;

  while (--l >= 0)
    {
      if ((*q = AE[*p++&0x7F]) == 0xFF)
	while (l-- >= 0)
	  *q++ = 0x40;
      ++q;
    }

  return( buf );
}


 static
 void
prunix( register struct IHADCB *dcbp, int c )
{
  int move = 0;

  switch (c)
    {
  case ' ':
    if(dcbp->access.qsam.Lcolpos < dcbp->lrecl)
      {
	++dcbp->access.qsam.Nspace;
	++dcbp->access.qsam.Lcolpos;
      }
    return;
  case '\b':
    if (dcbp->access.qsam.Lcolpos == 0)
      return;
    if (dcbp->access.qsam.Nspace > 0)
      {
	--dcbp->access.qsam.Nspace;
	--dcbp->access.qsam.Lcolpos;
	return;
      }
    move = -1;
  case '\t':
    break;
  case '\n':
    ++dcbp->access.qsam.LineNumber;
  case '\r':
  case '\f':
    dcbp->access.qsam.Lcolpos = 0;
    dcbp->access.qsam.Nspace = 0;
  default:
    move = (isprint(c) != 0);
    }

  if (dcbp->access.qsam.Lcolpos > 0 || move > 0)
    dcbp->access.qsam.Lcolpos += move;

  for ( ; dcbp->access.qsam.Nspace > 0; dcbp->access.qsam.Nspace -= 1)
    putc( ' ', dcbp->access.qsam.F );

  if (dcbp->access.qsam.Lcolpos <= dcbp->lrecl)
    putc( c, dcbp->access.qsam.F );
}



/*
 *  Interpret asa carriage control characters.  The first character
 *  of each line is assumed to be a control character; the meanings
 *  of the control characters are:
 *
 *  ' ' single-space before printing
 *  '0' double-space before printing
 *  '-' triple-space before printing
 *  '1' new page before printing
 *  '+' do not space at all before printing
 *
 *  A line beginning with '+' will overprint the previous line.
 *  Lines beginning with other than the above characters are
 *  treated as if they began with ' '.  Lines must be terminated
 *  with a null -- not a newline.
 */

#define	PAGELENGTH	66

 static
 void
prasa( register struct IHADCB *dcbp, register unsigned char *s )
{
  register int c;

  switch( *s++ )
    {

  /* new page */
  case '1':
    if (!dcbp->access.qsam.firstline)
      {
	while( dcbp->access.qsam.LineNumber < PAGELENGTH )
	  prunix( dcbp, '\n' );
	dcbp->access.qsam.LineNumber = 0;
      }
    break;

  /* triple space */
  case '-':
    prunix( dcbp, '\n' );
    if ( dcbp->access.qsam.LineNumber >= PAGELENGTH)
      dcbp->access.qsam.LineNumber = 0;
  /* double space */
  case '0':
    prunix( dcbp, '\n' );
    if ( dcbp->access.qsam.LineNumber >= PAGELENGTH)
      dcbp->access.qsam.LineNumber = 0;
  /* single space and invalid characters */
  default:
  case ' ':
    if (!dcbp->access.qsam.firstline)
      {
	prunix( dcbp, '\n' );
	if ( dcbp->access.qsam.LineNumber >= PAGELENGTH)
	  dcbp->access.qsam.LineNumber = 0;
      }
    break;

  /* no space at all */
  case '+':
    if (!dcbp->access.qsam.firstline)
      prunix( dcbp, '\r' );
    }

  /* remainder of line */
  while ( (c = *s++) )
    prunix( dcbp, c );

  dcbp->access.qsam.firstline = FALSE;
}


/*
 * OPEN
 */

struct IHADCB *remember[20];
boolean sigenabled = FALSE;
void (*oldsigint)(int);
static void bsin(), bsout(), bspoint();
static void qsin(), qsout();


 void
sigint( int sig )
{
  struct IHADCB *dcbp;
  int i;

  for( i = 0; i < DIM(remember); ++i )
    if ( (dcbp = remember[i]) )
      CLOSE( dcbp );

  if ( oldsigint != SIG_DFL && oldsigint != SIG_IGN )
    (*oldsigint)( sig );

  exit( 1 );
}
    

 void
OPEN( struct IHADCB *dcbp, int flag )
{
  char *type;
  const struct JFCB *ddp;
  int i;
  boolean isOutput;

  if ( (ddp=dcbp->jfcb) == NULL )
    {
      fprintf( stderr, "Invalid DCB\n" );
      abend( 0x00013000 );
    }

  if ( dcbp->flag )
    {
      fprintf( stderr, "//%s DD: DCB already open\n", ddp->ddname );
      abend( 0x00013000 );
    }

  if ( ddp->fname == NULL )
    {
#if 0
      fprintf( stderr, "//%s DD missing\n", ddp->ddname );
      abend( 0x00013000 );
#endif
      return;
    }

  dcbp->save.recfm = dcbp->recfm;
  dcbp->save.lrecl = dcbp->lrecl;

  /* override DCB from DD */
  if ( ddp->recfm != NA )
    dcbp->recfm = ddp->recfm;
  if ( ddp->lrecl > 0 )
    dcbp->lrecl = ddp->lrecl;

  if ( dcbp->exlst )
    (**(dcbp->exlst))( dcbp );

  if ( dcbp->dsorg != QSAM && dcbp->dsorg != BSAM )
    {
      fprintf( stderr, "//%s DD: DSORG is not BSAM or QSAM\n", ddp->ddname );
      abend( 0x00013000 );
    }
  if ( dcbp->macrf != O_RDONLY && dcbp->macrf != O_WRONLY && dcbp->macrf != O_RDWR )
    {
      fprintf( stderr, "//%s DD: MACRF is not O_RDONLY, O_WRONLY or O_RDWR\n",
	      ddp->ddname );
      abend( 0x00013000 );
    }

  switch( dcbp->oflag = flag )
    {
  case O_RDONLY:
      if ( dcbp->macrf == O_WRONLY )
	{
	  fprintf( stderr, "//%s DD: DCB cannot be opened for input\n", ddp->ddname );
	  abend( 0x00013000 );
	}

      if ( 0 == strcmp(ddp->fname,"-") )
	dcbp->fname = "stdin";
      else
	if ( ddp->fname[0] == '&' )
	  dcbp->fname = ddp->fname+1;
	else
	  {
	    if ( ddp->fname[0] == '|' )
	      {
		if ( dcbp->dsorg == BSAM )
		  {
		    fprintf( stderr, "//%s DD DSNAME=\"%s\",DSORG=BSAM: pipes not supported\n",
			    ddp->ddname, ddp->fname );
		    abend( 0x00013000 );
		  }
	      }
	    dcbp->fname = ddp->fname;
	  }
      break;
  case O_WRONLY:
      if ( dcbp->macrf == O_RDONLY )
	{
	  fprintf( stderr, "//%s DD: DCB cannot be opened for output\n", ddp->ddname );
	  abend( 0x00013000 );
	}
      dcbp->oflag |= O_CREAT | O_TRUNC;

      if ( 0 == strcmp(ddp->fname,"-") )
	dcbp->fname = "stdout";
      else
	if ( ddp->fname[0] == '&' )
	  dcbp->fname = ddp->fname+1;
	else
	  {
	    if ( ddp->fname[0] == '|' )
	      {
		if ( dcbp->dsorg == BSAM )
		  {
		    fprintf( stderr, "//%s DD DSNAME=\"%s\",DSORG=BSAM: pipes not supported\n",
			    ddp->ddname, ddp->fname );
		    abend( 0x00013000 );
		  }
	      }
	    dcbp->fname = ddp->fname;
	  }
      break;
  case O_RDWR:
      if ( dcbp->macrf != O_RDWR )
	{
	  fprintf( stderr, "//%s DD: DCB cannot be opened for update\n", ddp->ddname );
	  abend( 0x00013000 );
	}
      dcbp->oflag |= O_CREAT;

      if ( 0 == strcmp(ddp->fname,"-") )
	dcbp->fname = "/dev/tty";
      else
	if ( ddp->fname[0] == '&' )
	  dcbp->fname = ddp->fname+1;
	else
	  {
	    if ( ddp->fname[0] == '|' )
	      {
		if ( dcbp->dsorg == BSAM )
		  {
		    fprintf( stderr, "//%s DD DSNAME=\"%s\",DSORG=BSAM: pipes not supported\n",
			    ddp->ddname, ddp->fname );
		    abend( 0x00013000 );
		  }
	      }
	    dcbp->fname = ddp->fname;
	  }
      break;
    }

  if ( dcbp->recfm == NA )
    {
      fprintf( stderr, "//%s DD DSNAME=\"%s\": RECFM is required\n", ddp->ddname, dcbp->fname );
      abend( 0x00013000 );
    }
  if ( dcbp->lrecl == 0 )
    {
      fprintf( stderr, "//%s DD DSNAME=\"%s\": LRECL is required\n", ddp->ddname, dcbp->fname );
      abend( 0x00013000 );
    }

  switch( dcbp->dsorg )
    {
  case BSAM:
      if ( dcbp->recfm != F )
	{
	  fprintf( stderr, "//%s DD DSNAME=\"%s\",DSORG=BSAM: only RECFM=F is supported\n",
		  ddp->ddname, ddp->fname );
	  abend( 0x00013000 );
	}
      switch( dcbp->oflag & O_ACCMODE )
	{
      case O_RDONLY:
	  dcbp->in = bsin;
	  dcbp->out = NULL;
	  dcbp->access.bsam.point = bspoint;
	  break;
      case O_WRONLY:
	  dcbp->in = NULL;
	  dcbp->out = bsout;
	  dcbp->access.bsam.point = bspoint;
	  break;
      case O_RDWR:
	  dcbp->in = bsin;
	  dcbp->out = bsout;
	  dcbp->access.bsam.point = bspoint;
	  break;
	}
      if ( 0 == strcmp(dcbp->fname,"stdin") )
	dcbp->access.bsam.fd = fileno( stdin );
      else
	if ( 0 == strcmp(dcbp->fname,"stdout") )
	  dcbp->access.bsam.fd = fileno( stdin );
	else
	  if ( 0 == strcmp(dcbp->fname,"stderr") )
	    dcbp->access.bsam.fd = fileno( stderr );
	  else
	    if ( (dcbp->access.bsam.fd=open(dcbp->fname,dcbp->oflag,0666)) == -1 )
	      {
		char buffer[100];
		sprintf( buffer, "//%s DD DSNAME=\"%s\": cannot open", ddp->ddname, dcbp->fname );
		perror( buffer );
		abend( 0x00013000 );
	      }
      break;
  case QSAM:
      switch( dcbp->oflag & 0x03 )
	{
      case O_RDONLY:
	  if ( EBCDIC && (dcbp->recfm == V || dcbp->recfm == VA) )
	    {
	      fprintf( stderr, "//%s DD DSNAME=\"%s\": QSAM input is not supported for RECFM=V[A]\n",
		      ddp->ddname, dcbp->fname );
	      abend( 0x00013000 );
	    }
	  type = "r";
	  dcbp->in = qsin;
	  dcbp->out = NULL;
	  isOutput = FALSE;
	  break;
      case O_WRONLY:
	  if ( EBCDIC && (dcbp->recfm == V || dcbp->recfm == VA) )
	    {
	      fprintf( stderr, "//%s DD DSNAME=\"%s\": QSAM output is not supported for RECFM=V[A]\n",
		      ddp->ddname, dcbp->fname );
	      abend( 0x00013000 );
	    }
	  type = "w";
	  dcbp->in = NULL;
	  dcbp->out = qsout;
	  isOutput = TRUE;
	  break;
      case O_RDWR:
	  if ( EBCDIC && (dcbp->recfm == V || dcbp->recfm == VA) )
	    {
	      fprintf( stderr, "//%s DD DSNAME=\"%s\": QSAM input/output is not supported for RECFM=V[A]\n",
		      ddp->ddname, dcbp->fname );
	      abend( 0x00013000 );
	    }
	  if ( dcbp->fname[0] == '|' )
	    {
	      fprintf( stderr, "//%s DD DSNAME=\"%s\",DSORG=QSAM,MACRF=RW: pipes not supported\n",
		      ddp->ddname, dcbp->fname );
	      abend( 0x00013000 );
	    }
	  type = "r+";
	  dcbp->in = qsin;
	  dcbp->out = qsout;
	  isOutput = TRUE;
	  break;
	}
      if ( 0 == strcmp(dcbp->fname,"stdin") )
	dcbp->access.qsam.F = stdin;
      else
	if ( 0 == strcmp(dcbp->fname,"stdout") )
	  dcbp->access.qsam.F = stdout;
	else
	  if ( 0 == strcmp(dcbp->fname,"stderr") )
	    dcbp->access.qsam.F = stderr;
	  else
	    {
	      if ( dcbp->fname[0] != '|' )
		{
		  if ( (dcbp->access.qsam.F=fopen(dcbp->fname,type)) == NULL )
		    {
		      char buffer[100];
		      sprintf( buffer, "//%s DD DSNAME=\"%s\": cannot open",
			      ddp->ddname, dcbp->fname );
		      perror( buffer );
		      abend( 0x00013000 );
		    }
		}
	      else
		{
		  if ( (dcbp->access.qsam.F=popen(dcbp->fname+1,type)) == NULL )
		    {
		      char buffer[100];
		      sprintf( buffer, "//%s DD DSNAME=\"%s\": cannot open",
			      ddp->ddname, dcbp->fname );
		      perror( buffer );
		      abend( 0x00013000 );
		    }
		}
	    }
      if ( T && isOutput )
	setlinebuf( dcbp->access.qsam.F );
      dcbp->access.qsam.firstline = TRUE;
      dcbp->access.qsam.LineNumber = 0;
      dcbp->access.qsam.Lcolpos = 0;
      dcbp->access.qsam.Nspace = 0;
      break;
    }

  dcbp->flag = TRUE;

  if ( ! sigenabled )
    {
      sigenabled = TRUE;

      oldsigint = signal( SIGINT, sigint );
    }

  for( i = 0; i < DIM(remember); ++i )
    {
      if ( remember[i] == NULL )
	{
	  remember[i] = dcbp;
	  break;
	}
    }
}

 void
CLOSE( struct IHADCB *dcbp )
{
  const struct JFCB *ddp;
  int i;

  if ( dcbp->flag )
    {
      dcbp->flag = FALSE;

      switch( dcbp->dsorg )
	{
      case QSAM:
	  if ( dcbp->recfm==FA || dcbp->recfm==VA || dcbp->recfm==FAX || dcbp->recfm==VAX )
	    prunix( dcbp, '\n' );
	  if ( dcbp->access.qsam.F == stdin )
	    ;
	  else
	    if ( dcbp->access.qsam.F == stdout )
	      fflush( stdout );
	    else
	      if ( dcbp->access.qsam.F == stderr )
		fflush( stderr );
	      else
		{
		  if ( dcbp->fname[0] == '|' )
		    pclose( dcbp->access.qsam.F );
		  else
		    fclose( dcbp->access.qsam.F );
		}
	  break;
      case BSAM:
	  if ( dcbp->access.bsam.fd == fileno(stdin) ||
	       dcbp->access.bsam.fd == fileno(stdout) ||
	       dcbp->access.bsam.fd == fileno(stderr) )
	    ;
	  else
	    close( dcbp->access.bsam.fd );
	  break;
	}

      if ( (ddp=dcbp->jfcb)->fname[0] == '&' )
	unlink( ddp->fname+1 );

      dcbp->recfm = dcbp->save.recfm;
      dcbp->lrecl = dcbp->save.lrecl;
      dcbp->fname = NULL;
      dcbp->oflag = 0;

      for( i = 0; i < DIM(remember); ++i )
	{
	  if ( remember[i] == dcbp )
	    {
	      remember[i] = NULL;
	      break;
	    }
	}
    }
}


/*
 * BSAM read
 */
 static
 void
bsin( register struct IHADCB *dcbp, char *addr )
{
  int rc;
  const struct JFCB *ddp = dcbp->jfcb;

  if ( ! dcbp->flag )
    {
      fprintf( stderr, "//%s DD: DCB not opened properly\n", ddp->ddname );
      abend( 0x00001000 );
    }

  *(dcbp->eodad) = FALSE;
  errno = 0;

  if ( (rc=read(dcbp->access.bsam.fd,addr,dcbp->lrecl)) == 0 )
    {
      *(dcbp->eodad) = TRUE;
      return;
    }

  if ( rc == -1 || rc != dcbp->lrecl )
    (*(dcbp->synad))( dcbp, errno );
}


/*
 * BSAM point
 */
 static
 void
bspoint( register struct IHADCB *dcbp, unsigned int n )
{
  const struct JFCB *ddp = dcbp->jfcb;

  if ( ! dcbp->flag )
    {
      fprintf( stderr, "//%s DD: DCB not opened properly\n", ddp->ddname );
      abend( 0x00001000 );
    }

  errno = 0;

  if ( lseek(dcbp->access.bsam.fd,n*dcbp->lrecl,0) == -1 )
    (*(dcbp->synad))( dcbp, errno );
}


/*
 * BSAM write
 */
 static
 void
bsout( register struct IHADCB *dcbp, char *addr )
{
  const struct JFCB *ddp = dcbp->jfcb;

  if ( ! dcbp->flag )
    {
      fprintf( stderr, "//%s DD: DCB not opened properly\n", ddp->ddname );
      abend( 0x00001000 );
    }

  errno = 0;

  if ( write(dcbp->access.bsam.fd,addr,dcbp->lrecl) != dcbp->lrecl )
    (*(dcbp->synad))( dcbp, errno );
}


/*
 * QSAM read
 */
 static
 void
qsin( register struct IHADCB *dcbp, unsigned char *addr )
{
  unsigned char *p, buffer[BUFSIZ];
  const struct JFCB *ddp = dcbp->jfcb;
  int l;

  if ( ! dcbp->flag )
    {
      fprintf( stderr, "//%s DD: DCB not opened properly\n", ddp->ddname );
      abend( 0x00001000 );
    }

  *(dcbp->eodad) = FALSE;

  if ( EBCDIC )
    /* Do EBCDIC/ASCII translation for ..X recfm's and not for non-X recfm's */
    {
      if ( dcbp->recfm==FX || dcbp->recfm==VX || dcbp->recfm==FAX || dcbp->recfm==VAX )
	/* Translate from ASCII to EBCDIC */
	{
	  if ( fgets(buffer,BUFSIZ,dcbp->access.qsam.F) == NULL )
	    {
	      if ( ferror(dcbp->access.qsam.F) )
		(*(dcbp->synad))( dcbp, errno );
	      else
		*(dcbp->eodad) = TRUE;
	      return;
	    }

	  if ( (p=strchr(buffer,'\n')) == NULL )
	    {
	      fprintf( stderr, "//%s DD DSNAME=\"%s\": line too long\n",
		      ddp->ddname, dcbp->fname );
	      (*(dcbp->synad))( dcbp, 0 );
	      l = strlen( buffer );
	    }
	  else
	    {
	      *p = '\0';
	      l = p - buffer;
	    }

	  if ( dcbp->recfm==FX || dcbp->recfm==FAX )
	    {
	      if ( l > dcbp->lrecl )
		{
		  fprintf( stderr, "//%s DD DSNAME=\"%s\": line too long\n",
			  ddp->ddname, dcbp->fname );
		  (*(dcbp->synad))( dcbp, 0 );
		}
	      l = dcbp->lrecl;
	    }

	  if ( l > 0 )
	    {
	      p = tr_ae( buffer, l );

	      memcpy( addr, p, l );
	      addr[l] = '\0';
	    }
	  else
	    {
	      addr[0] = AE[' '];
	      addr[1] = '\0';
	    }
	}
      else
	/* Do not translate from ASCII to EBCDIC.  Input is IBM format.  */
	/* Recfm is either F or FA, since V[A] is not supported by OPEN() */
	{
	  if ( fread(addr,l=dcbp->lrecl,1,dcbp->access.qsam.F) != 1 ||
	       feof(dcbp->access.qsam.F) || ferror(dcbp->access.qsam.F) )
	    {
	      if ( ferror(dcbp->access.qsam.F) )
		(*(dcbp->synad))( dcbp, errno );
	      else
		*(dcbp->eodad) = TRUE;
	      return;
	    }

	  addr[l] = '\0';
	}
    }
  else
    /* No EBCDIC/ASCII translation is ever done.  All files are normal UNIX files */
    {
      if ( fgets(buffer,BUFSIZ,dcbp->access.qsam.F) == NULL )
	{
	  if ( ferror(dcbp->access.qsam.F) )
	    (*(dcbp->synad))( dcbp, errno );
	  else
	    *(dcbp->eodad) = TRUE;
	  return;
	}

      if ( (p=strchr(buffer,'\n')) == NULL )
	{
	  fprintf( stderr, "//%s DD DSNAME=\"%s\": line too long\n",
		    ddp->ddname, dcbp->fname );
	  (*(dcbp->synad))( dcbp, 0 );
	  l = strlen( buffer );
	}
      else
	{
	  *p = '\0';
	  l = p - buffer;
	}

      if ( dcbp->recfm==FX || dcbp->recfm==FAX || dcbp->recfm==F || dcbp->recfm==FA )
	{
	  if ( l > dcbp->lrecl )
	    {
	      fprintf( stderr, "//%s DD DSNAME=\"%s\": line too long\n",
			ddp->ddname, dcbp->fname );
	      (*(dcbp->synad))( dcbp, 0 );
	    }
	  l = dcbp->lrecl;
	}

      if ( l > 0 )
	{
	  strncpy( addr, buffer, l );
	  addr[l] = '\0';
	  
	  if ( l > strlen(addr) )
	    memset( addr+strlen(addr), ' ', l-strlen(addr) );
	}
      else
	{
	  addr[0] = ' ';
	  addr[1] = '\0';
	}
    }
}


/*
 * QSAM write
 */
 static
 void
qsout( register struct IHADCB *dcbp, char *addr )
{
  register int c;
  const struct JFCB *ddp = dcbp->jfcb;

  if ( ! dcbp->flag )
    {
      fprintf( stderr, "//%s DD: DCB not opened properly\n", ddp->ddname );
      abend( 0x00001000 );
    }

  if ( T )
    fflush( stderr );

  if ( EBCDIC )
    {
      switch( dcbp->recfm )
	{
	  int n;
      case FA:
	  if ( *addr == 0 )
	    addr = " ";
      case F:
	  for( n=dcbp->lrecl; n > 0 && (c = *addr++); --n )
	    putc( c, dcbp->access.qsam.F );
	  while( n-- > 0 )
	    putc( 0x40, dcbp->access.qsam.F );
	  break;
      case FX: case VX:
	  addr = tr_ea( addr );
	  while( (c = *addr++) )
	    prunix( dcbp, c );
	  prunix( dcbp, '\n' );
	  break;
      case FAX: case VAX:
	  addr = tr_ea( addr );
	  if ( *addr == 0 )
	    addr = " ";

	  prasa( dcbp, addr );
	  break;
	}
    }
  else
    {
      switch( dcbp->recfm )
	{
      case F: case V:
      case FX: case VX:
	  while( (c = *addr++) )
	    prunix( dcbp, c );
	  prunix( dcbp, '\n' );
	  break;
      case FA: case VA:
      case FAX: case VAX:
	  if ( *addr == 0 )
	    addr = " ";

	  prasa( dcbp, addr );
	  break;
	}
    }
}


/*
 * A ddname specification is:  NAME [ ( [RECFM] [,LRECL] ) ]
 *
 * If NAME begins with the character '&', the file will be deleted
 * when closed.  A NAME of "-" denotes stdin for input or stdout for
 * output.
 *
 * Supported access methods if EBCDIC is FALSE are (trailing X is
 * ignored):
 *
 *               +------- RECFM ------+
 *               |  F   FA    V   VA  |
 *     DSORG=QS _|____________________|
 *       Input   |  Y    Y    Y    Y  |
 *       Output  |  Y    Y    Y    Y  |
 *       Update  |  Y    Y    Y    Y  |
 *     DSORG=PS _|____________________|
 *       Input   |  Y    -    -    -  |
 *       Output  |  Y    -    -    -  |
 *       Update  |  Y    -    -    -  |
 *              -+--------------------+
 *
 * Supported access methods if EBCDIC is TRUE are:
 *
 *               +----------------- RECFM ----------------+
 *               |  F   FA   FX  FAX    V   VA   VX  VAX  |
 *     DSORG=QS _|________________________________________|
 *       Input   |  Y    Y    Y    Y    -    -    Y    Y  |
 *       Output  |  Y    Y    Y    Y    -    -    Y    Y  |
 *       Update  |  Y    Y    Y    Y    -    -    Y    Y  |
 *     DSORG=PS _|________________________________________|
 *       Input   |  Y    -    -    -    -    -    -    -  |
 *       Output  |  Y    -    -    -    -    -    -    -  |
 *       Update  |  Y    -    -    -    -    -    -    -  |
 *              -+----------------------------------------+
 */
char  *recfm[] ={ NULL,
		  "F",   "V",
		  "FA",  "VA",
		  "FX",  "VX",
		  "FAX", "VAX",
		};

 boolean
ddname( name, array, size )
  char *name;
  struct JFCB *array;
{
  char *p;
  int i, j;

  for( i=0; i<size; ++i )
    if ( array[i].fname == NULL )
      {
	array[i].fname = strtok(name,"(");
	array[i].recfm = NA;
	array[i].lrecl = 0;

	if ( (p=strtok(NULL,")")) != NULL )
	  {
	    while( *p == ' ' )
	      ++p;

	    if ( *p != ',' )
	      {
		p = ucase( strtok(p,",") );

		for( j=1; j<DIM(recfm); ++j )
		  if ( 0 == strcmp(p,recfm[j]) )
		    {
		      array[i].recfm = (enum recfm) j;
		      break;
		    }

		if ( 0 == strcmp(p,"FXA") )
		  array[i].recfm = FAX;
		if ( 0 == strcmp(p,"VXA") )
		  array[i].recfm = VAX;

		if ( array[i].recfm == NA )
		  {
		    fprintf( stderr, "//%s DD DSNAME=\"%s\": RECFM '%s' undefined\n",
			      array[i].ddname, array[i].fname, p );
		    array[i].fname = NULL;
		    return( FALSE );
		  }

		p = strtok( NULL, "" );
	      }
	    else
	      ++p;

	    if ( p != NULL )
	      array[i].lrecl = atoi( p );
	  }

	return( TRUE );
      }

  fprintf( stderr, "JFCB table overflow\n" );
  return( FALSE );
}


/*
 * Convert character string to upper case
 */
 char *
ucase( s )
  char *s;
{
  register char *p = s, c;

  if ( (p = s) )
    while( (c = *p) )
      {
	if ( isascii(c) && islower(c) )
	  *p = toupper( c );
	++p;
      }

  return( s );
}
